"""

This is the grid search part for the cmt inversion. Rise in the gradient sense

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

Last Update: April 2020

"""

import os
import glob
import yaml

# pycmt3d
from pycmt3d.source import CMTSource
from pycmt3d import DataContainer
from pycmt3d import WeightConfig
from pycmt3d import Gradient3dConfig, Gradient3d

# Get logger to log progress
from gcmt3d import logger


def read_yaml_file(filename):
    """read yaml file"""
    with open(filename) as fh:
        return yaml.load(fh, Loader=yaml.FullLoader)


def gradient(cmt_file_db, param_path):
    """Runs the actual inversion.

    :param cmt_file_db:
    :param param_path:
    :return: Nothing, inversion results are written to file.
    """

    # Inversion Params
    inversionparam_path = os.path.join(param_path,
                                       "CMTInversion/InversionParams.yml")
    INV_params = read_yaml_file(inversionparam_path)

    # Check if grid search should be performed from the parameters file
    if not bool(INV_params["gridsearch"]):
        return

    # Get processing path from cmt_filename in database
    cmt_dir = os.path.dirname(os.path.abspath(cmt_file_db))

    # Create cmt source:
    cmtsource = CMTSource.from_CMTSOLUTION_file(cmt_file_db)

    # Inversion dictionary directory
    inv_dict_dir = os.path.join(cmt_dir, "inversion", "inversion_dicts")

    # Inversion dictionaries
    inv_dict_files = glob.glob(os.path.join(inv_dict_dir, "grid*"))

    # Inversion output directory
    inv_out_dir = os.path.join(cmt_dir, "inversion", "inversion_output", "g3d")

    # WRite start of inversion process
    logger.info(" ")
    logger.info("#######################################################")
    logger.info("#                                                     #")
    logger.info("#      Starting inversion ...                         #")
    logger.info("#                                                     #")
    logger.info("#######################################################")
    logger.info(" ")

    # Creating Data container
    dcon = DataContainer()

    for _i, inv_dict_file in enumerate(inv_dict_files):

        # Get processing band
        bandstring1 = str(os.path.basename(inv_dict_file).split(".")[1])
        if "surface" in bandstring1 or "body" in bandstring1:
            bandstring = bandstring1.split("#")[0]
        else:
            bandstring = bandstring1

        band = [float(x) for x in bandstring.split("_")]

        logger.info(" ")
        logger.info("  " + 54 * "*")
        logger.info("  Getting data for inversion from period band:")
        logger.info("  Low: %d s || High: %d s" % tuple(band))
        logger.info("  " + 54 * "*")
        logger.info(" ")

        # Load inversion file dictionary
        inv_dict = read_yaml_file(inv_dict_file)
        asdf_dict = inv_dict["asdf_dict"]
        window_file = inv_dict["window_file"]

        # Adding measurements
        # Print Inversion parameters:
        logger.info("  Adding measurements to data container:")
        logger.info("  _____________________________________________________")
        logger.info(" ")

        # Add measurements from ASDF file and windowfile
        logger.info("  Window file:")
        logger.info("   " + window_file)
        logger.info(" ")
        logger.info("  ASDF files:")
        for key, value in asdf_dict.items():
            logger.info("     " + key + ": " + value)

        dcon.add_measurements_from_asdf(window_file, asdf_dict)

        logger.info("  _____________________________________________________")
        logger.info("  ... ")
        logger.info("  ")
        logger.info("  ... ")

    logger.info("  Setting up inversion classes .... ")
    logger.info("  " + 54 * "*")
    logger.info("  ... ")

    # Setting up weight config
    inv_weight_config = INV_params["weight_config"]

    grad3d_params = INV_params["grad3d_config"]

    weight_config_grad3d = WeightConfig(
        normalize_by_energy=inv_weight_config["normalize_by_energy"],
        normalize_by_category=inv_weight_config["normalize_by_category"],
        azi_bins=inv_weight_config["azi_bins"],
        azi_exp_idx=inv_weight_config["azi_exp_idx"])

    grad3d_config = Gradient3dConfig(
        method=grad3d_params["method"],
        weight_data=bool(grad3d_params["weight_data"]),
        weight_config=weight_config_grad3d,
        # flag to use the gradient method on inverted traces.
        taper_type=grad3d_params["taper_type"],
        c1=float(grad3d_params["c1"]),
        c2=float(grad3d_params["c2"]),
        idt=float(grad3d_params["idt"]),
        ia=float(grad3d_params["ia"]),
        nt=int(grad3d_params["nt"]),
        nls=int(grad3d_params["nls"]),
        crit=float(grad3d_params["crit"]),
        precond=bool(grad3d_params["precond"]),
        reg=bool(grad3d_params["reg"]),
        bootstrap=bool(grad3d_params["bootstrap"]),
        bootstrap_repeat=int(grad3d_params["bootstrap_repeat"]),
        bootstrap_subset_ratio=float(grad3d_params["bootstrap_subset_ratio"]),
        mpi_env=bool(grad3d_params["mpi_env"]),
        parallel=bool(grad3d_params["parallel"]))

    logger.info("  PyCMT3D is finding an improved CMTSOLUTION .... ")
    logger.info("  " + 54 * "*")
    logger.info(" ")
    logger.info(" ")

    # Create Gradient3d class
    g3d = Gradient3d(cmtsource, dcon, grad3d_config)

    # Run inversion
    g3d.search()

    # Plot results
    if bool(INV_params["summary_plot"]):
        pass
        # g3d.plot_summary(inv_out_dir, figure_format="pdf")

    if bool(INV_params["statistics_plot"]):
        # Plot Statistics for inversion
        g3d.plot_stats_histogram(outputdir=inv_out_dir)

    if bool(INV_params["summary_json"]):
        g3d.write_summary_json(outputdir=inv_out_dir, mode="global")

    if bool(INV_params["write_new_cmt"]):
        g3d.write_new_cmtfile(outputdir=inv_out_dir)

    if bool(INV_params["write_new_synt"]):
        g3d.write_new_syn(outputdir=os.path.join(inv_out_dir, "new_synt"),
                          file_format="asdf")

    if bool(INV_params["plot_new_synthetics"]):
        g3d.plot_new_synt_seismograms(outputdir=os.path.join(inv_out_dir,
                                                             "waveform_plots"),
                                      figure_format="pdf")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', action='store', dest='cmt_file',
                        required=True, help="Path to CMT file in database")
    parser.add_argument('-p', action='store', dest='param_path', required=True,
                        help="Path to Parameter Directory")
    args = parser.parse_args()

    gradient(args.cmt_file, args.param_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This script will create the perturbed Moment tensors in the perturbation
directories.


:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)

Last Update: January 2020
"""

from gcmt3d.stats.stats import Statistics
import argparse


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('dirname', help='Database directory', type=str)
    parser.add_argument('-d', action="store_true", dest='direct',
                        help='Direct directory')
    parser.add_argument('-o', dest='outdir', required=False, default='./',
                        help='Output directory', type=str)
    args = parser.parse_args()

    # Load shit
    ST = Statistics._from_dir(args.dirname, direct=args.direct)

    # Plot it
    ST.plot_changes(savedir=args.outdir)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Methods that handles Window selection

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""
import os
import pyflex
import obspy
import copy
import importlib


def plot_window_figure(figure_dir, figure_id, ws, _verbose=False,
                       figure_format="pdf"):
    """
    Plot window figure out

    :param figure_dir: output figure directory
    :type figure_dir: str
    :param figure_id: figure id to distinguish windows plots, for
        example, trace id could be used, like "II.AAK.00.BHZ"
    :type figure_id: str
    :param ws: window selector object from pyflex
    :type ws: pyflex.WindowSelector
    :param _verbose: verbose output flag
    :type _verbose: bool
    :param figure_format: figure format, could be "pdf", "png" and etc.
    :type figure_format: str
    :return:
    """
    outfn = "%s.%s" % (figure_id, figure_format)
    figfn = os.path.join(figure_dir, outfn)
    if _verbose:
        print("Output window figure:", figfn)
    ws.plot(figfn)


def update_user_levels(user_module, config, station, event, obsd, synt):
    """Update user levels as an array using user_module

    This function emulates the user_functions feature in FLEXWIN. User
    has to prepare a python module that has a method called
    generate_user_levels. This module has to be passed out as a string
    (It should be formatted as you would in an import statement). And
    then, user_module is dynamically imported and the
    generate_user_levels function is called with the following
    parameters: pyflex config, obspy station data, obspy event data,
    observed and synthetic obspy traces. It is excepted that user to
    use given scalar acceptance levels as base values and generate
    acceptance level arrays as long as trace data to be given to the
    pyflex. Function should return these values in order: stalta water
    level, tshift acceptance level, dlna acceptance level,
    cc_acceptance_level, s2n_limit.

    :param user_module: user module as a string
    :type user_module: str
    :param config: window selection config
    :type config_dict: pyflex.Config
    :param station: station information which provids station location to
        calculate the epicenter distance
    :type station: obspy.Inventory or pyflex.Station
    :param event: event information, providing the event information
    :type event: pyflex.Event, obspy.Catalog or obspy.Event
    :param observed: observed trace
    :type observed: obspy.Trace
    :param synthetic: synthetic trace
    :type synthetic: obspy.Trace
    :return: window selection config with arrays
    :rtype: pyflex.Config
    """
    # Ridvan Orsvuran, 2016
    # If user gives a user_module, allow it to create the user
    # acceptance levels as arrays.
    try:
        # Add current working directory to path. This enables user to
        # use python files in the working directory.
        # import sys
        # sys.path.append(".")

        # Import the user module
        user = importlib.import_module(user_module)
        # Assign generate function to a variable. This enables to
        # catch the AttributeError.
        generate_user_levels = user.generate_user_levels
    except ImportError:
        raise Exception("Could not import the user_function module: %s"
                        % user_module)
    except AttributeError:
        raise Exception("Given user module does not have a "
                        "generate_user_levels method: %s" % user_module)

    # do not give the original config to the user
    config_copy = copy.deepcopy(config)
    stalta_waterlevel, tshift, dlna, cc, s2n = generate_user_levels(
        config_copy, station, event, obsd, synt)

    # Create a new config using new acceptance levels
    new_config = copy.deepcopy(config)
    new_config.stalta_waterlevel = stalta_waterlevel
    new_config.tshift_acceptance_level = tshift
    new_config.dlna_acceptance_level = dlna
    new_config.cc_acceptance_level = cc
    new_config.s2n_limit = s2n

    return new_config


def window_on_trace(obs_tr, syn_tr, config, station=None,
                    event=None, user_module=None, _verbose=False,
                    figure_mode=False, figure_dir=None,
                    figure_format="pdf"):
    """
    Window selection on a trace(obspy.Trace)

    :param observed: observed trace
    :type observed: obspy.Trace
    :param synthetic: synthetic trace
    :type synthetic: obspy.Trace
    :param config: window selection config
    :type config_dict: pyflex.Config
    :param station: station information which provids station location to
        calculate the epicenter distance
    :type station: obspy.Inventory or pyflex.Station
    :param event: event information, providing the event information
    :type event: pyflex.Event, obspy.Catalog or obspy.Event
    :param user_module: user module as a string
    :type user_module: str
    :param figure_mode: output figure flag
    :type figure_mode: bool
    :param figure_dir: output figure directory
    :type figure_dir: str
    :param _verbose: verbose flag
    :type _verbose: bool
    :return:
    """

    if not isinstance(obs_tr, obspy.Trace):
        raise ValueError("Input obs_tr should be obspy.Trace")
    if not isinstance(syn_tr, obspy.Trace):
        raise ValueError("Input syn_tr should be obspy.Trace")
    if not isinstance(config, pyflex.Config):
        raise ValueError("Input config should be pyflex.Config")

    # Ridvan Orsvuran, 2016
    # If user gives a user_module, use it to update acceptance levels
    # as arrays.
    if user_module is not None and user_module != "None":
        config = update_user_levels(user_module, config, station,
                                    event, obs_tr, syn_tr)

    ws = pyflex.WindowSelector(obs_tr, syn_tr, config,
                               event=event, station=station)
    try:
        windows = ws.select_windows()
    except Exception as err:
        print("Error(%s): %s" % (obs_tr.id, err))
        windows = []

    if figure_mode:
        plot_window_figure(figure_dir, obs_tr.id, ws, _verbose,
                           figure_format=figure_format)

    if _verbose:
        print("Station %s picked %i windows" % (obs_tr.id, len(windows)))

    return windows


def window_on_stream(observed, synthetic, config_dict, station=None,
                     event=None, user_modules=None,
                     figure_mode=False, figure_dir=None,
                     _verbose=False):
    """
    Window selection on a Stream

    :param observed: observed stream
    :type observed: obspy.Stream
    :param synthetic: synthetic stream
    :type synthetic: obspy.Stream
    :param config_dict: window selection config dictionary, for example,
        {"Z": pyflex.Config, "R": pyflex.Config, "T": pyflex.Config}
    :type config_dict: dict
    :param station: station information which provids station location to
        calculate the epicenter distance
    :type station: obspy.Inventory or pyflex.Station
    :param event: event information, providing the event information
    :type event: pyflex.Event, obspy.Catalog or obspy.Event
    :param user_modules: user_module strings in a dict similar to config_dict.
    :type user_modules: dict
    :param figure_mode: output figure flag
    :type figure_mode: bool
    :param figure_dir: output figure directory
    :type figure_dir: str
    :param _verbose: verbose flag
    :type _verbose: bool
    :return:
    """
    if not isinstance(observed, obspy.Stream):
        raise ValueError("Input observed should be obspy.Stream")
    if not isinstance(synthetic, obspy.Stream):
        raise ValueError("Input synthetic should be obspy.Stream")
    if not isinstance(config_dict, dict):
        raise ValueError("Input config_dict should be dict")

    all_windows = {}

    # Ridvan Orsvuran, 2016
    # Assign an empty dict to user_modules if it is None to avoid errors.
    if user_modules is None:
        user_modules = {}

    for category in config_dict:
        config_base = config_dict[category]
        user_module = user_modules.get(category, None)
        if len(category) == 1:
            # then it is component
            obs = observed.select(component=category)
        elif len(category) == 3:
            # then it is channel
            obs = observed.select(channel=category)
        else:
            raise ValueError("The length of Config_dict.keys()[%s] should be "
                             "either 1 or 3, for example, ['E', 'N', 'Z'] "
                             "or ['BHE', 'BHN', 'BHZ']" % config_dict.keys())

        for obs_tr in obs:
            component = obs_tr.stats.channel[-1]
            try:
                syn_tr = synthetic.select(station=obs_tr.stats.station,
                                          network=obs_tr.stats.network,
                                          component=component)[0]
            except Exception as err:
                print("Couldn't find corresponding synt for obsd trace(%s):"
                      "%s" % (obs_tr.id, err))
                continue

            config = copy.deepcopy(config_base)
            windows = window_on_trace(
                obs_tr, syn_tr, config, station=station,
                event=event, user_module=user_module, _verbose=_verbose,
                figure_mode=figure_mode, figure_dir=figure_dir)

            if windows is None:
                continue

            # Notice: Ebru suggests to write out window even its length is
            # zero, which means no windows selected on the traces, in order
            # to keep track of every thing
            all_windows[obs_tr.id] = windows

    return all_windows



#!/usr/bin/env python
"""
Scripts that contains methods that deals with station information in the
asdf file.

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)
"""
from __future__ import (print_function, division, absolute_import)
import pyasdf
from ..station import extract_staxml_info


def extract_station_info_from_asdf(asdf, verbose=False):
    """ extract the sensor type from stationxml in asdf file """
    if isinstance(asdf, str) or isinstance(asdf, chr):
        ds = pyasdf.ASDFDataSet(asdf, mode='r')
    elif isinstance(asdf, pyasdf.ASDFDataSet):
        ds = asdf
    else:
        raise TypeError("Input asdf either be a filename or "
                        "pyasdf.ASDFDataSet")

    asdf_sensors = dict()
    ntotal = len(ds.waveforms)
    for idx, st_group in enumerate(ds.waveforms):
        if verbose:
            print("[%4d/%d]Station: %s"
                  % (idx, ntotal, st_group._station_name))
        try:
            inv = st_group.StationXML
            info = extract_staxml_info(inv)
            asdf_sensors.update(info)
        except Exception as msg:
            print("Failed to extract due to: %s" % msg)
            continue

    print("Number of stations and channels: %d, %d"
          % (ntotal, len(asdf_sensors)))

    return asdf_sensors


def extract_waveform_stations(asdf, stations=None):
    """
    Extract station information from wavefrom group
    """
    if isinstance(asdf, str):
        ds = pyasdf.ASDFDataSet(asdf, mode='r')
    elif isinstance(asdf, pyasdf.ASDFDataSet):
        ds = asdf
    else:
        raise TypeError("Input asdf either be a filename or "
                        "pyasdf.ASDFDataSet")

    sta_dict = {}
    if stations is None:
        stations = ds.waveforms.list()
    for st_id in stations:
        station_group = getattr(ds.waveforms, st_id)
        if "StationXML" not in dir(station_group):
            continue
        staxml = getattr(station_group, "StationXML")
        sta_dict[st_id] = [staxml[0][0].latitude, staxml[0][0].longitude,
                           staxml[0][0].elevation,
                           staxml[0][0][0].depth]

    return sta_dict


def extract_adjoint_stations(asdf, stations=None):
    """
    Extract station information from adjoint source group
    """
    if isinstance(asdf, str):
        ds = pyasdf.ASDFDataSet(asdf, mode='r')
    elif isinstance(asdf, pyasdf.ASDFDataSet):
        ds = asdf
    else:
        raise TypeError("Input asdf either be a filename or "
                        "pyasdf.ASDFDataSet")

    sta_dict = {}
    try:
        adjsrcs = ds.auxiliary_data.AdjointSources
    except Exception:
        return {}

    if stations is None:
        stations = adjsrcs.list()
    for adj_name in stations:
        adj = getattr(adjsrcs, adj_name)
        pars = adj.parameters
        station_id = pars["station_id"]
        if station_id not in sta_dict:
            sta_dict[station_id] = [pars["latitude"], pars["longitude"],
                                    pars["elevation_in_m"],
                                    pars["depth_in_m"]]

    return sta_dict



"""

This script contains functions to convolve the Green's functions in a Stream
with a source time function.

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

Last Update: November 2019

"""

import numpy as np


def gaussian(t, to, sig, amp):
    """Create Gaussian pulse.

    :param t: time vector
    :param to: time in the center
    :param sig: standard deviation
    :param amp: amplitude
    :return: vector with same size of t containing the corresponding pulse

    """
    return amp * np.exp(-0.5 * (t-to)**2 / (sig * sig))


def bimodal_gaussian(t, amp1=2.5, to1=20, sig1=5, amp2=1.75, to2=32, sig2=3):
    """ Creates bimodal gaussian from gaussian function.

    :param t: time vector
    :param amp1: amplitude of first gaussian
    :param to1: center time of first gaussian
    :param sig1: standard deviation of first gaussian
    :param amp2: amplitude of second gaussian
    :param to2: center time of second gaussian
    :param sig2: standard deviation of second gaussian
    :return:
    """

    g1 = gaussian(t, to1, sig1, amp1)
    g2 = gaussian(t, to2, sig2, amp2)

    return g1 + g2


# class Wavelet(object):
#     """Contains wavelet"""
#
#     def __init__(self, t, data=None, type='gaussian', **kwargs):
#         """ Creates wavelet with parameters
#
#         :param dt: sampling time
#         :param N:
#         :param type: type of automatically created wavelet. Irrelevant if
#                      data is input.
#         :param
#         :param **kwargs: keyword arguments depending on the type of wavelet.
#                          'gaussian' or 'bimodal-gaussian'
#
#         """
#
#         self.t = t
#
#         if data is not None:
#             self.data = data
#             self.type = 'self-defined'
#         else:
#             if type == 'gaussian':
#                 self.wavefun = gaussian
#                 self.data = self.wavefun(t, *kwargs)
#
#     @classmethod
#     def _from_SAC(cls, filename):
#         """ Creates
#
#         :param filename: filename to create wavelet from
#         :type: str
#         :return:
#
#         """
#         pass
#
#     @classmethod
#     def _from_cmtsource(cmtsource, dt):
#         """CMTSource object provides halfduration and time shift,
#         as well as approximation of the Gaussian.
#
#         From Specfem Manual:
#
#             The zero time of the simulation corresponds to the center
#             of the triangle/Gaussian, or the centroid time of the
#             earthquake. The start time of the simulation is
#             t = −1.5 ∗ half duration (the 1.5 is to make sure the
#             moment rate function is very close to zero when starting
#             the simulation). To convert to absolute time tabs, set
#
#                 t_abs = tpde + time shift + tsynthetic
#
#             where tpde is the time given in the first line of
#             the CMTSOLUTION, time shift is the corresponding
#             value from the original CMTSOLUTION file and tsynthetic
#             is the time in the first column of the output seismogram.
#
#         Arguments:
#             cmtsource (CMTSource object): Input CMTSource object inlcuding
#                                           necessary info such as the half
#                                           duration, time shift, and, of
#                                           course, origin time.
#
#         Returns:
#             Wavelet class: [description]
#         """
#
#         return cls()


# def convolve_trace(tr: obspy.trace, wavelet):
#     """Takes in a wavelet and an obspy Trace object and convolves the
#     wavelet with the Trace objectss
#
#     Args:
#         tr (obspy.Trace): Trace object
#         wavelet ([type]): wavelet
#     """
#     pass
#



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Methods that handles rotation of seismograms as extension to Obspy.
It can rotate `12`, `EN` and `RT`, forward and backward.

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""
from __future__ import print_function
import numpy as np
from obspy.geodetics import gps2dist_azimuth
from obspy import Stream
from .rotate_utils import check_orthogonality, rotate_12_rt


def calculate_baz(elat, elon, slat, slon):
    _, _, baz = gps2dist_azimuth(elat, elon, slat, slon)
    return baz


def ensemble_synthetic_channel_orientation(chan):
    """
    For synthetic data from SPECFEM, just ensemble the orientation
    information.
    """
    orientation = {"Z": (90, 0),
                   "N": (0, 0),
                   "E": (0, 90)}
    try:
        dip, azi = orientation[chan[-1]]
    except KeyError:
        raise Exception("Orientation is not defined for synthetics"
                        " of component: %s" % chan)
    return dip, azi


def extract_channel_orientation(tr, inv):
    """ Extract the dip and azimuth from inventory, given the trace """
    try:
        nw = tr.stats.network
        sta = tr.stats.station
        loc = tr.stats.location
        chan = tr.stats.channel

        if loc == "S3":
            # If channel is synthetic(from specfem), return default value
            # for dip and azimuth.
            dip, azi = ensemble_synthetic_channel_orientation(chan)
        else:
            chan_inv = inv.select(network=nw, station=sta, location=loc,
                                  channel=chan)[0][0][0]
            dip, azi = chan_inv.dip, chan_inv.azimuth
    except Exception as errmsg:
        print("Unable to extract channel orientation information [%s]"
              " due to: %s" % (tr.id, errmsg))
        dip, azi = None, None

    return dip, azi


def extract_station_location(st, inventory):
    """
    Extract the station latitude and longitude from inventory, given
    stream.
    """
    nw = st[0].stats.network
    station = st[0].stats.station
    _inv = inventory.select(network=nw, station=station)
    sta_lat = float(_inv[0][0].latitude)
    sta_lon = float(_inv[0][0].longitude)
    return sta_lat, sta_lon


def check_vertical_inventory_sanity(tr, inventory):
    """
    Check the inventory of vertical(Z) component, check
    if the abs(dip) is 90 and azimuth is 0.
    """
    if tr.stats.channel[-1] != "Z":
        raise ValueError("Function only checks vertical(Z) component(%s)"
                         % tr.stats.channel)
    dip, azi = extract_channel_orientation(tr, inventory)

    if dip is None or azi is None:
        return False

    if np.isclose(abs(dip), 90.0) and np.isclose(abs(azi), 0.0):
        return True
    else:
        return False


def check_horizontal_inventory_sanity(tr1, tr2, inventory):
    """
    Check two horizontal components and see if their dip is 0
    and azimuth is orthogonal to each other.

    :param tr1:
    :param tr2:
    :param inventory:
    :return:
    """
    if tr1.id[:-1] != tr2.id[:-1]:
        raise ValueError("Two horizontal ids should share the same network,"
                         "station, location and channel[0:2]: %s, %s"
                         % (tr1.id, tr2.id))

    if tr1.stats.channel[-1] == "Z" or tr2.stats.channel[-1] == "Z":
        raise ValueError("Functions should check two horizontal component:"
                         "%s, %s" % (tr1.id, tr2.id))
    dip1, azi1 = extract_channel_orientation(tr1, inventory)
    dip2, azi2 = extract_channel_orientation(tr2, inventory)

    if dip1 is None or azi1 is None or dip2 is None or azi2 is None:
        return False

    # check dip
    if not np.isclose(dip1, 0.0) or not np.isclose(dip2, 0.0):
        return False

    # check azimuth
    if not check_orthogonality(azi1, azi2):
        return False

    return True


def check_information_before_rotation(i_1, i_2, inv, sanity_check=False):
    # check starttime, sampling rate
    dt = 0.5 * i_1.stats.delta
    if (len(i_1) != len(i_2)) or \
            (abs(i_1.stats.starttime - i_2.stats.starttime) > dt) \
            or (i_1.stats.sampling_rate != i_2.stats.sampling_rate):
        msg = "All components need to have the same time span."
        raise ValueError(msg)

    # check inventory sanity if required by user
    if sanity_check:
        if not check_horizontal_inventory_sanity(i_1, i_2, inv):
            msg = "Horizontal component are not orthogonal to " \
                  "each other: %s, %s" % (i_1.id, i_2.id)
            raise ValueError(msg)


def rotate_12_rt_func(st, inv, back_azimuth, method="12->RT",
                      sanity_check=False):
    """
    Rotate horizontal component to RT. This function works generally
    for two horizontal and orthogonal components. This function
    supports two method:

    1. "12->RT": which rotates "12" component to "RT", for example,
       "BH1" and "BH2" to "BHR" and "BHT"
    2. "NE->RT": which rotates "NE" component to "RT", for example,
       "BHN" and "BHE" to "BHR" and "BHT"

    The reason why we use our own rotation function is because in obspy
    the inventory information is not checked.

    :param st: input stream
    :param inv: station inventory information
    :param method: rotation method
    :param back_azimuth: back azimuth(station to event azimuth)
    :return: rotated stream
    """
    if method not in ["12->RT", "NE->RT"]:
        raise ValueError("rotate_12_rt_func only supports method:"
                         "['12->RT', 'NE->RT']")

    bad_ids = []
    input_components, output_components = method.split("->")
    if len(input_components) == 2:
        input_1 = st.select(component=input_components[0])
        input_2 = st.select(component=input_components[1])
        for i_1, i_2 in zip(input_1, input_2):
            try:
                check_information_before_rotation(
                    i_1, i_2, inv, sanity_check=sanity_check)
            except Exception as err:
                bad_ids.extend([i_1.id, i_2.id])
                print("Unable to rotate [%s, %s] due to: %s"
                      % (i_1.id, i_2.id, err))
                continue
            inc1, azi1 = extract_channel_orientation(i_1, inv)
            inc2, azi2 = extract_channel_orientation(i_2, inv)
            if azi1 is None or azi2 is None:
                continue
            output_1, output_2 = rotate_12_rt(i_1.data, i_2.data, back_azimuth,
                                              azi1, azi2)
            if output_1 is None or output_2 is None:
                continue
            i_1.data = output_1
            i_2.data = output_2
            # Rename the components
            i_1.stats.channel = i_1.stats.channel[:-1] + output_components[0]
            i_2.stats.channel = i_2.stats.channel[:-1] + output_components[1]
            # Add the azimuth back to stats object
            for comp in (i_1, i_2):
                comp.stats.back_azimuth = back_azimuth
    else:
        raise ValueError("Wrong method: %s" % method)

    # remove trace in bad_ids
    # for bid in bad_ids:
    #    st.remove(st.select(id=bid)[0])

    return st


def rotate_rt_to_ne(st, baz):
    """ Use obspy rotate function to rotate from RT to NE """
    st.rotate(method="RT->NE", back_azimuth=baz)


def remove_bad_z_component(st, inventory):
    """ remove Z component if its inventory does not meet requirements """
    for tr in st:
        if tr.stats.channel[-1] == "Z":
            if not check_vertical_inventory_sanity(tr, inventory):
                st.remove(tr)


def rotate_to_rt(st, baz, inventory, mode, sanity_check=False):
    """
    Rotate horizontal components(12 or NE) to RT directions.

    :param st:
    :param baz:
    :param inventory:
    :param mode:
    :param sanity_check:
    :return:
    """
    components = [tr.stats.channel[-1] for tr in st]

    if sanity_check:
        # remove bad Z component
        remove_bad_z_component(st, inventory)

    if mode in ["12->RT", "ALL->RT"]:
        if "1" in components and "2" in components:
            try:
                rotate_12_rt_func(st, inventory, method="12->RT",
                                  back_azimuth=baz)
            except Exception as errmsg:
                print("Error rotating 12->RT:%s" % errmsg)

    if mode in ["NE->RT", "ALL->RT"]:
        if "N" in components and "E" in components:
            try:
                rotate_12_rt_func(st, inventory, baz, method="NE->RT",
                                  sanity_check=sanity_check)
            except Exception as e:
                print("Error rotating NE->RT:%s" % e)


def rotate_one_station_stream(st, event_latitude, event_longitude,
                              station_latitude=None, station_longitude=None,
                              inventory=None, mode="NE->RT",
                              sanity_check=False):
    """
    Rotate the stream from the same network, station, location and channel
    code, for example the stream should only contains traces whose ids are
    "II.AAK.00.BH*"
    """

    mode = mode.upper()
    mode_options = ["NE->RT", "ALL->RT", "12->RT", "RT->NE"]
    if mode not in mode_options:
        raise ValueError("rotate_stream mode(%s) should be within %s"
                         % (mode, mode_options))
    if mode[0:2] == "RT" and sanity_check:
        raise ValueError("if rotating from RT to NE, then set "
                         "sanity_check to False")

    if inventory is None and \
            (station_latitude is None and station_longitude is None):
        raise ValueError("You need either specify station inventory "
                         "or station location information")

    if mode in ["12->RT", "ALL->RT"] and inventory is None:
        raise ValueError("If you want to rotate 12, you need to specify "
                         "station inventory")

    if len(sort_stream_by_station(st)) != 1:
        raise ValueError("This method only accepts stream that contains "
                         "one station")

    if station_latitude is None or station_longitude is None:
        try:
            station_latitude, station_longitude = \
                extract_station_location(st, inventory)
        except Exception as errmsg:
            # station_latitude and station_longitude is unknown
            # so the stream should be skipped
            print("Error extracting staiton latitude and longitude from "
                  "staiton inventory: %s" % errmsg)
            return

    baz = calculate_baz(event_latitude, event_longitude,
                        station_latitude, station_longitude)

    if mode in ["NE->RT", "12->RT", "ALL->RT"]:
        rotate_to_rt(st, baz, inventory, mode, sanity_check=sanity_check)

    if mode in ["RT->NE"]:
        rotate_rt_to_ne(st, baz)

    return st


def sort_stream_by_station(st):
    """
    Sort the traces in stream. Group the traces from same network,
    station, location, and channel code together
    """
    ntotal = len(st)
    sta_dict = {}
    n_added = 0
    for tr in st:
        nw = tr.stats.network
        station = tr.stats.station
        loc = tr.stats.location
        chan = tr.stats.channel[0:2]
        station_id = "%s.%s.%s.%s" % (nw, station, loc, chan)
        if station_id not in sta_dict:
            sta_dict[station_id] = \
                st.select(network=nw, station=station,
                          location=loc, channel="%s*" % chan)
            n_added += len(sta_dict[station_id])

    if n_added != ntotal:
        raise ValueError("Sort stream by station errors: number of traces "
                         "is inconsistent")
    return sta_dict


def rotate_stream(st, event_latitude, event_longitude,
                  inventory, mode="ALL->RT", sanity_check=False):
    """
    Rotate a stream to radial and transverse components based on the
    station information and event information

    :param st: input stream
    :type st: obspy.Stream
    :param event_latitude: event latitude
    :type event_latitude: float
    :param event_longitude: event longitude
    :type event_longitude: float
    :param inventory: station inventory information. If you want to rotate
        "12" components, you need to provide inventory since only station
        and station_longitude is not enough.
    :type inventory: obspy.Inventory
    :param mode: rotation mode, could be one of:

        1. "NE->RT": rotate only North and East channel to RT
        2. "12->RT": rotate only 1 and 2 channel, like "BH1" and "BH2" to RT
        3. "ALL->RT": rotate all components to RT
        4. "RT->NE": rotate RT to NE

    :param sanity_check: check the sanity of inventory, mianly of the
        orientation of ZNE components.

        1. If rotating observed data from NE(12)to RT, it is recommended
           to set to True; if rotating from RT to NE, then you could set
           it to False;
           ATTENTION: please turn it to True when processing observed
           data since the instrument alignment of observed data might
           be messy.
        2. If rotating synthetic data, you could set it to False since
           I assume for synthetic data, there should not be problem
           associated with orientation.

    :return: rotated stream(obspy.Stream)
    """

    rotated_stream = Stream()

    mode = mode.upper()
    mode_options = ["NE->RT", "ALL->RT", "12->RT", "RT->NE"]
    if mode not in mode_options:
        raise ValueError("rotate_stream mode(%s) should be within %s"
                         % (mode, mode_options))
    if mode[0:2] == "RT" and sanity_check:
        raise ValueError("if rotating from RT to NE, then set "
                         "sanity_check to False")

    if mode in ["12->RT", "ALL->RT"] and inventory is None:
        raise ValueError("Mode %s required inventory(stationxml) "
                         "information provided(to rotate '12')" % mode)

    # the stream might contains multiple stations so each station should
    # be rotated independently, or stations with different channel,
    # (for example, BH and EH), and different locations(such as
    # "00" and "10").
    sorted_st_dict = sort_stream_by_station(st)

    for sta_stream in sorted_st_dict.values():
        # loop over stations
        nw = sta_stream[0].stats.network
        station = sta_stream[0].stats.station
        loc = sta_stream[0].stats.location
        chan = sta_stream[0].stats.channel

        if loc == "S3" or chan[0:2] == "MX":
            # SPECFEM HACK: if the synthetic is generated by SPECFEM
            # and the stationxml is from read data, then the location
            # won't match between synt and staxml
            station_inv = inventory.select(network=nw, station=station)
        else:
            station_inv = inventory.select(network=nw, station=station,
                                           location=loc)

        _st = \
            rotate_one_station_stream(sta_stream, event_latitude,
                                      event_longitude, inventory=station_inv,
                                      mode=mode, sanity_check=sanity_check)
        if _st is not None:
            rotated_stream += _st

    return rotated_stream



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This script will download the observed data. To the necessary places.


:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""

from gcmt3d.utils.download import read_station_file
from gcmt3d.asdf.utils import smart_read_yaml, is_mpi_env
from gcmt3d.source import CMTSource
import os
from obspy.clients.fdsn.mass_downloader import RectangularDomain, \
    Restrictions, MassDownloader
import logging
from ..log_util import modify_logger

logger = logging.getLogger(__name__)
modify_logger(logger)


def data_request(cmt_filename, param_path):

    # Request config_file
    request_param_path = os.path.join(param_path,
                                      "RequestParams/RequestParams.yml")

    # Read the parameter file
    rCparams = smart_read_yaml(request_param_path, mpi_mode=is_mpi_env())

    # Earthquake and Station parameters
    cmt_dir = os.path.dirname(cmt_filename)
    station_dir = os.path.join(cmt_dir, "station_data")

    # Get STATIONS file from CMT directory
    stationsfile = os.path.join(station_dir, "STATIONS")

    # Observed output dir
    obsd_dir = os.path.join(cmt_dir, "seismograms", "obs")

    # CMT parameter input
    cmt = CMTSource.from_CMTSOLUTION_file(cmt_filename)
    duration = rCparams['duration']
    starttime_offset = rCparams['starttime_offset']

    starttime = cmt.origin_time + starttime_offset
    endtime = starttime + duration

    # Get station_list from station_file in database entry
    stations = read_station_file(stationsfile)

    # Create list of networks to download from
    networks = list(set([station[0] for station in stations]))
    network_string = ",".join(networks)

    # Set domain containing all locations
    # Rectangular domain containing parts of southern Germany.
    domain = RectangularDomain(minlatitude=-90, maxlatitude=90,
                               minlongitude=-180, maxlongitude=180)

    # Set download restrictions
    restrictions = Restrictions(
        starttime=starttime,
        endtime=endtime,
        reject_channels_with_gaps=False,
        minimum_length=float(rCparams['minimum_length']),
        # Trace needs to be almost full length
        network=network_string,  # Only certain networks
        channel=",".join(rCparams['channels']),
        location=",".join(rCparams['locations']))

    # No specified providers will result in all known ones being queried.
    providers = ["IRIS"]
    mdl = MassDownloader(providers=providers)
    # The data will be downloaded to the ``./waveforms/`` and ``./stations/``
    # folders with automatically chosen file n
    stationxml_storage = os.path.join(station_dir)
    waveform_storage = os.path.join(obsd_dir)
    logger.info("MSEEDs: %s" % waveform_storage)
    logger.info("XMLs: %s" % stationxml_storage)

    mdl.download(domain, restrictions, mseed_storage=waveform_storage,
                 stationxml_storage=stationxml_storage)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Methods that contains utils for adjoint sources

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""
from __future__ import (absolute_import, division, print_function)
import collections
from obspy import UTCDateTime
from obspy.core.inventory import Channel, Station, Network, Inventory, Site


def check_in_range(value, vranges):
    if vranges[0] > vranges[1]:
        vmin = vranges[1]
        vmax = vranges[0]
    else:
        vmin = vranges[0]
        vmax = vranges[1]

    if value < vmin or value > vmax:
        raise ValueError("Value(%f) not in range: %s" % (value, vranges))


def write_stations_file(sta_dict, filename="STATIONS"):
    """
    Write station information out to a txt file(in SPECFEM FORMAT)

    :param sta_dict: the dict contains station locations information.
        The key should be "network.station", like "II.AAK".
        The value are the list of
        [latitude, longitude, elevation_in_m, depth_in_m].
    :type sta_dict: dict
    :param filename: the output filename for STATIONS file.
    :type filename: str
    """
    with open(filename, 'w') as fh:
        od = collections.OrderedDict(sorted(sta_dict.items()))
        for _sta_id, _sta in od.items():
            network, station = _sta_id.split(".")
            _lat = _sta[0]
            _lon = _sta[1]
            check_in_range(_lat, [-90.1, 90.1])
            check_in_range(_lon, [-180.1, 180.1])
            fh.write("%-9s %5s %15.4f %12.4f %10.1f %6.1f\n"
                     % (station, network, _lat, _lon, _sta[2], _sta[3]))


def create_simple_inventory(network, station, latitude=None, longitude=None,
                            elevation=None, depth=None, start_date=None,
                            end_date=None, location_code="S3",
                            channel_code="MX"):
    """
    Create simple inventory with only location information,
    for ZNE component, especially usefull for synthetic data
    """
    azi_dict = {"MXZ": 0.0,  "MXN": 0.0, "MXE": 90.0}
    dip_dict = {"MXZ": 90.0, "MXN": 0.0, "MXE": 0.0}
    channel_list = []

    if start_date is None:
        start_date = UTCDateTime(0)

    # specfem default channel code is MX
    for _comp in ["Z", "E", "N"]:
        _chan_code = "%s%s" % (channel_code, _comp)
        chan = Channel(_chan_code, location_code, latitude=latitude,
                       longitude=longitude, elevation=elevation,
                       depth=depth, azimuth=azi_dict[_chan_code],
                       dip=dip_dict[_chan_code], start_date=start_date,
                       end_date=end_date)
        channel_list.append(chan)

    site = Site("N/A")
    sta = Station(station, latitude=latitude, longitude=longitude,
                  elevation=elevation, channels=channel_list, site=site,
                  creation_date=start_date, total_number_of_channels=3,
                  selected_number_of_channels=3)

    nw = Network(network, stations=[sta, ], total_number_of_stations=1,
                 selected_number_of_stations=1)

    inv = Inventory([nw, ], source="SPECFEM3D_GLOBE", sender="Princeton",
                    created=UTCDateTime.now())

    return inv



"""

Plot utilities not to modify plots or base plots.

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

Last Update: April 2020


"""


import matplotlib.pyplot as plt
import numpy as np
import matplotlib
from matplotlib import cm
from matplotlib import colors
from matplotlib.colors import ListedColormap, LinearSegmentedColormap


params = {
    'font.weight': 'bold',
    'axes.labelweight': 'bold',
    'axes.labelsize': 9,
    'xtick.labelsize': 7,
    'xtick.direction': 'in',
    'xtick.top': True,   # draw label on the top
    'xtick.bottom': True,    # draw label on the bottom
    'xtick.minor.visible': True,
    'xtick.major.top': True,  # draw x axis top major ticks
    'xtick.major.bottom': True,  # draw x axis bottom major ticks
    'xtick.minor.top': True,  # draw x axis top minor ticks
    'xtick.minor.bottom': True,  # draw x axis bottom minor ticks
    'ytick.labelsize': 7,
    'ytick.direction': 'in',
    'ytick.left': True,  # draw label on the top
    'ytick.right': True,  # draw label on the bottom
    'ytick.minor.visible': True,
    'ytick.major.left': True,  # draw x axis top major ticks
    'ytick.major.right': True,  # draw x axis bottom major ticks
    'ytick.minor.left': True,  # draw x axis top minor ticks
    'ytick.minor.right': True,  # draw x axis bottom minor ticks
    # 'text.usetex': True,
    # 'font.family': 'STIXGeneral',
    # 'mathtext.fontset': 'cm',
}
matplotlib.rcParams.update(params)


def remove_topright(ax=None):
    """Removes top and right border and ticks from input axes."""

    # Get current axis if none given.
    if ax is None:
        ax = plt.gca()

    # Hide the right and top spines
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

    # Only show ticks on the left and bottom spines
    ax.yaxis.set_ticks_position('left')
    ax.xaxis.set_ticks_position('bottom')


def remove_all(ax=None, top=False, bottom=False, left=False, right=False,
               xticks='none', yticks='none'):
    """Removes all frames and ticks."""
    # Get current axis if none given.
    if ax is None:
        ax = plt.gca()

    # Hide the right and top spines
    ax.spines['bottom'].set_visible(bottom)
    ax.spines['left'].set_visible(left)
    ax.spines['right'].set_visible(right)
    ax.spines['top'].set_visible(top)

    # Only show ticks on the left and bottom spines
    ax.yaxis.set_ticks_position(yticks)
    ax.xaxis.set_ticks_position(xticks)

    # Turn off tick labels
    ax.set_yticklabels([])
    ax.set_xticklabels([])


def get_color(x, cmap="seismic", vmin=0, vmax=100, norm=None):
    n = vmax - vmin
    val = (np.clip(x, vmin, vmax) - vmin) / n

    if type(cmap) not in [ListedColormap, LinearSegmentedColormap]:
        cmap = getattr(cm, cmap)

    if norm is not None:
        val = norm(val)

    return cmap(val)


def sns_to_mpl(sns_cmap, linear=False):
    """Converts sns colormap to matplotlib colormap. If the ``as_cmap``
    flag is on for seaborn cmap a linear list is returned. we don't want
    that."""
    if linear:
        return LinearSegmentedColormap(sns_cmap.as_hex())
    else:
        return ListedColormap(sns_cmap.as_hex())


def create_colorbar(vmin, vmax, cmap="seismic", norm=None, cax=None, **kwargs):
    """Creates Colorbar given certain inputs."""

    if type(cmap) not in [ListedColormap, LinearSegmentedColormap]:
        cmap = getattr(cm, cmap)

    if norm is not None:
        values = np.linspace(vmin, vmax, cmap.N)
        norm = colors.BoundaryNorm(values, cmap.N)
    else:
        norm = colors.Normalize(vmin=vmin, vmax=vmax)
    sm = cm.ScalarMappable(norm=norm, cmap=cmap)

    ax = plt.gca()
    if cax is None:
        c = plt.colorbar(sm, ax=ax, **kwargs)
    else:
        c = plt.colorbar(sm, cax=cax, **kwargs)
    return c



"""

This file contains functions to plot statistics

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

Last Update: November 2019

"""


import os
import logging
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.gridspec import GridSpec
from cartopy.crs import PlateCarree
import cartopy
from obspy.imaging.beachball import beach
import matplotlib
from matplotlib import cm
from matplotlib import colors
from scipy.odr import RealData, ODR, Model
import seaborn as sns

from .plot_util import remove_topright, remove_all
from .plot_util import create_colorbar
from .plot_util import sns_to_mpl
from .plot_util import get_color

from ..log_util import modify_logger

logger = logging.getLogger(__name__)
modify_logger(logger)

params = {
    'font.weight': 'bold',
    'axes.labelweight': 'bold',
    'axes.labelsize': 9,
    'xtick.labelsize': 7,
    'xtick.direction': 'in',
    'xtick.top': True,  # draw label on the top
    'xtick.bottom': True,  # draw label on the bottom
    'xtick.minor.visible': True,
    'xtick.major.top': True,  # draw x axis top major ticks
    'xtick.major.bottom': True,  # draw x axis bottom major ticks
    'xtick.minor.top': True,  # draw x axis top minor ticks
    'xtick.minor.bottom': True,  # draw x axis bottom minor ticks
    'ytick.labelsize': 7,
    'ytick.direction': 'in',
    'ytick.left': True,  # draw label on the top
    'ytick.right': True,  # draw label on the bottom
    'ytick.minor.visible': True,
    'ytick.major.left': True,  # draw x axis top major ticks
    'ytick.major.right': True,  # draw x axis bottom major ticks
    'ytick.minor.left': True,  # draw x axis top minor ticks
    'ytick.minor.right': True,  # draw x axis bottom minor ticks
    # 'text.usetex': True,
    # 'font.family': 'STIXGeneral',
    # 'mathtext.fontset': 'cm',
}
matplotlib.rcParams.update(params)


# Define a function (quadratic in our case) to fit the data with.
def linear_func(p, x):
    m, c = p
    return m*x + c


def fit_xy(x, y):
    # Create a model for fitting.
    linear_model = Model(linear_func)

    # Create a RealData object using our initiated data from above.
    data = RealData(x, y)

    # Set up ODR with the model and data.
    odr = ODR(data, linear_model, beta0=[0., 1.])

    # Run the regression.
    out = odr.run()

    return out.beta


def format_exponent(ax, axis='y'):

    # Change the ticklabel format to scientific format
    ax.ticklabel_format(axis=axis, style='sci', scilimits=(-2, 2))

    # Get the appropriate axis
    if axis == 'y':
        ax_axis = ax.yaxis
        x_pos = 0.0
        y_pos = 1.0
        horizontalalignment = 'left'
        verticalalignment = 'bottom'
    else:
        ax_axis = ax.xaxis
        x_pos = 1.0
        y_pos = -0.05
        horizontalalignment = 'right'
        verticalalignment = 'top'

    # Run plt.tight_layout() because otherwise the offset text doesn't update
    plt.tight_layout()
    # THIS IS A BUG
    # Well, at least it's sub-optimal because you might not
    # want to use tight_layout(). If anyone has a better way of
    # ensuring the offset text is updated appropriately
    # please comment!

    # Get the offset value
    offset = ax_axis.get_offset_text().get_text()

    if len(offset) > 0:
        # Get that exponent value and change it into latex format
        minus_sign = u'\u2212'
        expo = np.float(offset.replace(minus_sign, '-').split('e')[-1])
        offset_text = '$\\times\\mathregular{10^{%d}}$' % expo

        # Turn off the offset text that's calculated automatically
        ax_axis.offsetText.set_visible(False)

        # Add in a text box at the top of the y axis
        ax.text(x_pos, y_pos, offset_text, transform=ax.transAxes,
                horizontalalignment=horizontalalignment,
                verticalalignment=verticalalignment)
    return ax


def plot_cmts(ax, latitude, longitude, depth, mt, nmmt, alpha):

    for (lon, lat, d, m, sm) \
            in zip(longitude.tolist(), latitude.tolist(), depth.tolist(),
                   mt.tolist(), nmmt.tolist()):
        try:
            b = beach(m, linewidth=0.25, facecolor='k', bgcolor='w',
                      edgecolor='k', alpha=alpha, xy=(lon, lat), width=20,
                      size=100, nofill=False, zorder=100,
                      axes=ax)

            ax.add_collection(b)
        except Exception as e:
            print(e)


class PlotStats(object):
    """Plots statistics given the necessary variables."""

    def __init__(self, ocmt=None, ncmt=None, dCMT=None, xcorr_mat=None,
                 mean_mat=None, std_mat=None, stat_dict=None, labels=None,
                 dlabels=None, stations=None, nbins=20, npar=9, savedir=None):
        """
        Parameters:
            ocmt: old cmt matrix
            ncmt: new cmt matrix
            dCMT: diff cmt matrix
            xcorr_mat: cross correlation matrix
            mean_mat: mean vector
            std_mat: std vector
            labels: labels
            dlabels: delta label
            nbins: bins in the histograms
            savedir: directory to save the figure output
            verbose: verbosity

        The matrices below should have following columns:
            M0, Mrr, Mtt, Mpp, Mrt, Mrp, Mtp,
            depth, lat, lon, CMT, hdur, t_shift

        Station list rows have following content
            [network station latitude longitude elevation]
        """

        self.ocmt = ocmt
        self.depth = ocmt[:, 7]
        self.latitude = ocmt[:, 8]
        self.longitude = ocmt[:, 9]
        self.N = ocmt.shape[0]
        self.ncmt = ncmt
        self.dCMT = dCMT
        self.xcorr_mat = xcorr_mat
        self.mean_mat = mean_mat
        self.std_mat = std_mat
        self.labels = labels
        self.dlabels = dlabels
        self.stations = stations
        self.stat_dict = stat_dict

        # Save destination
        self.savedir = savedir

        # Fix depth
        self.ocmt[:, 7] = self.ocmt[:, 7]/1000
        self.ncmt[:, 7] = self.ncmt[:, 7]/1000
        self.dCMT[:, 7] = self.dCMT[:, 7]/1000
        self.mean_mat[7] = self.mean_mat[7]/1000
        self.std_mat[7] = self.std_mat[7]/1000
        # Min max ddepth for cmt plotting
        self.maxddepth = np.max(self.dCMT[:, 7])
        self.minddepth = np.min(self.dCMT[:, 7])
        self.dd_absmax = np.max(np.abs(
            [np.quantile(np.min(self.dCMT[:, 7]), 0.30),
             np.quantile(np.min(self.dCMT[:, 7]), 0.70)]))
        self.maxdepth = np.max(self.ocmt[:, 7])
        self.mindepth = np.min(self.ocmt[:, 7])

        self.nbins = nbins
        self.dmbins = np.linspace(-0.5, 0.5 + 0.5 / self.nbins, self.nbins)
        self.ddegbins = np.linspace(-0.1, 0.1 + 0.1 / self.nbins, self.nbins)
        self.dzbins = np.linspace(-25, 25 + 25 / self.nbins, self.nbins)
        self.dtbins = np.linspace(-10, 10 + 10 / self.nbins, self.nbins)

        # Map characteristics
        self.cl = 180.0  # central_longitude
        self.cmt_cmap = sns_to_mpl(
            sns.color_palette("colorblind", n_colors=5),
            linear=False)
        self.depth_cmap = "plasma"
        # self.cmt_cmap = dj.mpl_colormap

        # Measurement label and tag dictionary
        self.vtype_dict = {r'Time-shift: ${\Delta t}_{CC}$': "tshift",
                           r'$CC_{max} = _{max}|\frac{(d \star s)(\tau)}'
                           r'{\sqrt{(d_i^2) * (s_i^2)}}|$': "cc",
                           r'$P{L_1} = 10\log\left(\frac{d_i}'
                           r'{s_i}\right)$ [dB]':
                               "power_l1",
                           r'$P{L_2} = 10\log\left(\frac{d_i^2}'
                           r'{s_i^2}\right)$ [dB]':
                               "power_l2",
                           r'$P_{CC} = 10\log\frac{(d_i s_i)}'
                           r'{s_i^2}$ [dB]': "cc_amp",
                           r'$\frac{1}{2}\left|  d_i - s_i \right|^2$': "chi"}

    def plot_main_stats(self):
        """Plots summary of the main statistics"""

        # Create figure handle
        fig = plt.figure(figsize=(11, 6))

        # Create subplot layout
        GS = GridSpec(3, 4)

        cbar_dict = {"orientation": "horizontal",
                     "shrink": 0.5,
                     "fraction": 0.025,
                     "pad": 0.025}

        # Create axis for map
        ax = fig.add_subplot(GS[:2, :2],
                             projection=PlateCarree(central_longitude=self.cl))
        self.plot_map()
        self.plot_cmts()
        c = create_colorbar(vmin=-self.dd_absmax,
                            vmax=self.dd_absmax,
                            cmap=self.cmt_cmap,
                            norm=None, **cbar_dict)
        c.set_label(r'Change in Depth [km]')
        fontsize = 9
        text_dict = {"fontsize": fontsize,
                     "verticalalignment": 'top',
                     "backgroundcolor": "white",
                     "bbox": {"facecolor": 'white',
                              "edgecolor": "black"},
                     "zorder": 100}

        ax.text(0.005, 0.995, "%d EQs" % self.N,
                **text_dict, horizontalalignment='left',
                transform=ax.transAxes)

        # Create axis for map
        ax = fig.add_subplot(GS[:2, 2:],
                             projection=PlateCarree(central_longitude=self.cl))
        self.plot_map()
        self.plot_stations()
        # This is only done, so that both maps have the same aspect ratio
        c = create_colorbar(vmin=-self.dd_absmax,
                            vmax=self.dd_absmax,
                            cmap=self.cmt_cmap,
                            norm=None, **cbar_dict)
        c.ax.set_visible(False)

        ax.text(0.995, 0.995, "%d Stations" % len(self.stations),
                **text_dict, horizontalalignment='right',
                transform=ax.transAxes)

        # Change of parameter as function of depth
        msize = 15

        # Depth vs. change in depth
        ax = fig.add_subplot(GS[2:3, 0])
        plt.scatter(self.dCMT[:, 7], self.ocmt[:, 7],
                    c=get_color(self.dCMT[:, 7],
                                vmin=-self.dd_absmax,
                                vmax=self.dd_absmax,
                                cmap=self.cmt_cmap),
                    s=msize, marker='s', alpha=0.9,
                    )
        plt.plot([0, 0], [0, np.max(self.ocmt[:, 7])],
                 "k--", lw=1.5)
        plt.ylim(([0, np.max(self.ocmt[:, 7])]))
        plt.xlim(([np.min(self.dCMT[:, 7]), np.max(self.dCMT[:, 7])]))
        ax.invert_yaxis()
        plt.xlabel(r"Depth Change [km]")
        plt.ylabel(r"Depth [km]")

        fig.add_subplot(GS[2, 1])
        self.plot_histogram(self.dCMT[:, 7], self.nbins)
        remove_topright()
        plt.xlabel(r"Depth Change [km]")
        plt.ylabel("$N$", rotation=0, horizontalalignment='right')

        fig.add_subplot(GS[2, 2])
        self.plot_histogram(self.dCMT[:, -1], self.nbins)
        remove_topright()
        plt.xlabel(r"Origin Time Change [sec]")
        plt.ylabel(r"$N$", rotation=0, horizontalalignment='right')

        fig.add_subplot(GS[2, 3])
        self.plot_histogram(self.dCMT[:, 0]*100, self.nbins)
        remove_topright()
        plt.xlabel(r"Scalar Moment Change [%]")
        plt.ylabel(r"$N$", rotation=0, horizontalalignment='right')

        # Finally plot shot
        plt.tight_layout(pad=2, w_pad=2.5, h_pad=2.25)
        if self.savedir is not None:
            plt.savefig(os.path.join(self.savedir, "main_stats.pdf"))
        else:
            plt.show()

    @staticmethod
    def plot_scatter_hist(x, y, nbins, z=None, cmap=None, histc='grey',
                          zmin=None, zmax=None):
        """

        :param x: Data type x-axis
        :param y: Data type y-axis
        :params nbins: number of bins
        :param z: Data type to color xy datapoints. Default None. Datapoints in
                  color of histogram.
        :param cmap: name
        :param zmin: Colorbar min
        :param zmin: Colorbar max

        :return:
        """

        # definitions for the axes
        left, width = 0.125, 0.65
        bottom, height = 0.125, 0.65
        spacing = 0.000

        # Create dimensions
        rect_scatter = [left, bottom, width, height]
        rect_histx = [left, bottom + height + spacing, width, 0.15]
        rect_histy = [left + width + spacing, bottom, 0.15, height]

        # Create Axes
        ax_scatter = plt.axes(rect_scatter)
        ax_scatter.tick_params(direction='in', top=True, right=True)
        ax_histx = plt.axes(rect_histx)
        ax_histx.tick_params(direction='in', labelbottom=False)
        ax_histy = plt.axes(rect_histy)
        ax_histy.tick_params(direction='in', labelleft=False)

        # scatterplot with color
        if cmap is not None and z is not None:
            # the scatter plot:
            if zmin is not None:
                vminz = zmin
            else:
                vminz = np.min(z)

            if zmax is not None:
                vmaxz = zmax
            else:
                vmaxz = np.max(z)

            ax_scatter.scatter(x, y, c=get_color(z, vmin=vminz,
                                                 vmax=vmaxz, cmap=cmap),
                               s=15, marker='s')

            # Colorbar
            cbar_dict = {"orientation": "horizontal"}
            cax = ax_scatter.inset_axes([0.05, 0.03, 0.25, 0.03],
                                        zorder=100)
            create_colorbar(vminz, vmaxz, cmap=cmap, cax=cax, **cbar_dict)
            cax.tick_params(left=False, right=False, bottom=True, top=True,
                            labelleft=False, labelright=False, labeltop=True,
                            labelbottom=False, which='both',
                            labelsize=6)

        # scatterplot without color
        else:
            # the scatter plot:
            ax_scatter.scatter(x, y, c=histc, s=15, marker='s')
            cax = None

        # Write out correlation coefficient in the top right
        corr_coeff = np.corrcoef(x, y)
        text_dict = {"fontsize": 6, "verticalalignment": 'top', "zorder": 100}
        ax_scatter.text(0.97, 0.97, "R = %1.2f" % corr_coeff[0, 1],
                        horizontalalignment='right', **text_dict,
                        transform=ax_scatter.transAxes)

        # now determine nice limits by hand:
        ax_scatter.set_xlim((np.min(x), np.max(x)))
        ax_scatter.set_ylim((np.min(y), np.max(y)))

        # Histogram settings
        binsx = np.linspace(np.min(x), np.max(x), nbins + 1)
        binsy = np.linspace(np.min(y), np.max(y), nbins + 1)
        ax_histx.hist(x, bins=binsx,
                      color=histc, ec=None)
        ax_histx.set_xlim(ax_scatter.get_xlim())
        ax_histy.hist(y, bins=binsy, orientation='horizontal',
                      color=histc, ec=None)
        ax_histy.set_ylim(ax_scatter.get_ylim())

        # Remove boundaries
        remove_all(ax=ax_histx, bottom=True)
        remove_all(ax=ax_histy, left=True)

        if cax is not None:
            return ax_scatter, ax_histx, ax_histy, cax
        else:
            return ax_scatter, ax_histx, ax_histy

    def plot_dM_dz(self):
        """Creates Figure with histograms one the side for two
        change in depth and change in scalar moment."""

        # start with a rectangular Figure
        plt.figure(figsize=(4, 4))

        ax_scatter, ax_histx, ax_histy, cax = self.plot_scatter_hist(
            self.dCMT[:, 0] * 100, self.dCMT[:, 7], self.nbins,
            z=self.ocmt[:, 7], cmap=self.depth_cmap, histc='grey',
            zmin=np.quantile(self.ocmt[:, 7], .0),
            zmax=np.quantile(self.ocmt[:, 7], .975))
        ax_scatter.set_ylabel(r"Origin Time Change [sec]")
        ax_scatter.set_xlabel(r"Scalar Moment Change [%]")
        ax_scatter.plot([np.min(self.dCMT[:, 0]) * 100,
                         np.max(self.dCMT[:, 0]) * 100],
                        [0, 0], 'k', zorder=-1, lw=0.75)
        ax_scatter.plot([0, 0], [np.min(self.dCMT[:, 7]),
                                 np.max(self.dCMT[:, 7])],
                        'k', zorder=0.1, lw=0.75)

        # Finally plot shot
        if self.savedir is not None:
            plt.savefig(os.path.join(self.savedir, "dM_dz.pdf"))
        else:
            plt.show()

    def plot_dM_z(self):
        """Creates Figure with histograms one the side for two
        change in depth and change in scalar moment."""

        # start with a rectangular Figure
        plt.figure(figsize=(4, 4))

        ax_scatter, ax_histx, ax_histy, cax = self.plot_scatter_hist(
            self.dCMT[:, 0] * 100, self.ocmt[:, 7], self.nbins,
            z=self.dCMT[:, 7], cmap=self.cmt_cmap, histc='grey',
            zmin=-self.dd_absmax, zmax=self.dd_absmax)

        ax_scatter.set_ylabel(r"Depth [km]")
        ax_scatter.set_xlabel(r"Scalar Moment Change [%]")
        ax_scatter.set_yscale('log')
        ax_scatter.set_ylim([10, self.maxdepth])
        ax_scatter.invert_yaxis()
        ax_histy.set_yscale(ax_scatter.get_yscale())
        ax_histy.set_ylim(ax_scatter.get_ylim())

        # Finally plot shot
        if self.savedir is not None:
            plt.savefig(os.path.join(self.savedir, "dM_z.pdf"))
        else:
            plt.show()

    def plot_changes(self):
        """Plots figure with statistics."""

        # Create figure handle
        fig = plt.figure(figsize=(11, 10))

        # Create subplot layout
        GS = GridSpec(6, 6)

        # Create axis for map
        fig.add_subplot(GS[:2, :3],
                        projection=PlateCarree(central_longitude=180.0))
        self.plot_map()
        self.plot_cmts()
        plt.title("Inversion statistics for %d earthquakes" % (self.N))

        # Create axis for map
        fig.add_subplot(GS[2:4, :3],
                        projection=PlateCarree(central_longitude=180.0))
        self.plot_map()
        self.plot_stations()
        plt.title("Stations used in the inversion")

        # table axes
        # fig.add_subplot(GS[2:4, 3])
        # self.plot_table()

        # change in cmttime
        fig.add_subplot(GS[2, 3])
        self.plot_histogram(self.dCMT[:, 10], self.dtbins)
        plt.xlabel(r"$\delta t$")

        # MT
        counter = 1
        for _i in range(2):
            for _j in range(3):
                fig.add_subplot(GS[0 + _i, 3 + _j])
                self.plot_histogram(self.dCMT[:, counter],
                                    self.dmbins, facecolor=(0.8, 0.8, 0.8))
                plt.xlabel("%s" % (self.dlabels[counter]))
                counter += 1

        # loc_ax
        fig.add_subplot(GS[3, 4])
        self.plot_histogram(self.dCMT[:, 8], self.ddegbins)
        plt.xlabel("$\\delta$Lat [$^{\\circ}$]")
        fig.add_subplot(GS[3, 5])
        self.plot_histogram(self.dCMT[:, 9], self.ddegbins)
        plt.xlabel("$\\delta$Lon [$^{\\circ}$]")
        fig.add_subplot(GS[2, 4])
        self.plot_histogram(self.dCMT[:, 7], self.dzbins)
        plt.xlabel("$\\delta z$ [km]")
        fig.add_subplot(GS[2, 5])
        self.plot_histogram(self.dCMT[:, 0], self.dmbins)
        plt.xlabel("$\\delta M_0$")

        vmin = 10**(25.75)
        vmax = 10**(26.5)
        msize = 10

        # Change of parameter as function of depth
        fig.add_subplot(GS[4:, 0:2])  # moment vs depth
        sc = plt.scatter(self.dCMT[:, 0], self.ocmt[:, 7], c=self.ocmt[:, 0],
                         s=msize, marker='o', cmap=cm.rainbow,
                         norm=colors.LogNorm(vmin=vmin, vmax=vmax))
        cbar = plt.colorbar(sc, orientation="horizontal")
        cbar.ax.set_ylabel(r"$M_0$")
        plt.xlim([-0.5, 0.5])
        plt.gca().invert_yaxis()
        plt.xlabel("$\\delta M_0$")
        plt.ylabel("$z$ [km]")

        # ddepth vs depth
        fig.add_subplot(GS[4:, 2:4])
        sc1 = plt.scatter(self.dCMT[:, 7], self.ocmt[:, 7], c=self.ocmt[:, 0],
                          s=msize, marker='o', cmap=cm.rainbow,
                          norm=colors.LogNorm(vmin=vmin, vmax=vmax))
        cbar = plt.colorbar(sc1, orientation="horizontal")
        cbar.ax.set_ylabel(r"$M_0$")
        plt.xlim([-20, 10])
        plt.xlabel(r"$\delta z$ [km]")
        plt.ylabel(r"$z$ [km]")
        plt.gca().invert_yaxis()

        # ddepth vs dM0
        fig.add_subplot(GS[4:, 4:])
        sc2 = plt.scatter(self.dCMT[:, 0], self.dCMT[:, 7], c=self.ocmt[:, 7],
                          s=msize, marker='o', cmap=cm.rainbow,
                          norm=colors.LogNorm())
        cbar = plt.colorbar(sc2, orientation="horizontal")
        cbar.ax.set_ylabel(r"$z$ [km]")
        plt.xlim([-0.5, 0.5])
        plt.ylim([-20, 10])
        plt.ylabel(r"$\delta z$ [km]")
        plt.xlabel(r"$\delta M_0$")
        plt.gca().invert_yaxis()

        # Finally plot shot
        plt.tight_layout()
        if self.savedir is not None:
            plt.savefig(os.path.join(self.savedir, "statfigure.pdf"))
        else:
            plt.show()

    def plot_map(self):

        ax = plt.gca()
        ax.set_global()
        ax.frameon = True
        ax.outline_patch.set_linewidth(0.75)

        # Set gridlines. NO LABELS HERE, there is a bug in the gridlines
        # function around 180deg
        gl = ax.gridlines(crs=PlateCarree(central_longitude=180.0),
                          draw_labels=False,
                          linewidth=1, color='lightgray', alpha=0.5,
                          linestyle='-', zorder=-1.5)
        gl.xlabels_top = False
        gl.ylabels_left = False
        gl.xlines = True

        # Add Coastline
        ax.add_feature(cartopy.feature.LAND, zorder=-2, edgecolor='black',
                       linewidth=0.5, facecolor=(0.9, 0.9, 0.9))

    def plot_cmts(self):

        ax = plt.gca()
        for idx, (lon, lat, m) in enumerate(zip(self.longitude, self.latitude,
                                                self.ncmt[:, 1:7])):
            try:
                # Longitude fix because cartopy is being shitty
                if self.cl == 180.0:
                    if lon <= 0:
                        lon = lon + 180.0
                    else:
                        lon = lon - 180.0
                b = beach(m, linewidth=0.25,
                          facecolor=get_color(self.dCMT[idx, 7],
                                              cmap=self.cmt_cmap,
                                              vmin=-self.dd_absmax,
                                              vmax=self.dd_absmax),
                          bgcolor='w',
                          edgecolor='k', alpha=1,
                          xy=(lon, lat), width=10,
                          size=10, nofill=False, zorder=-1,
                          axes=ax,
                          )

                ax.add_collection(b)
            except Exception as e:
                for line in e.__str__().splitlines():
                    logger.error(line)

    def plot_stations(self):
        """Plots stations into a map
        """

        slat = [station[0] for station in self.stations]
        # Weird fix because cartopy is weird
        if self.cl == 180.0:
            slon = [station[1] + self.cl if station[1] <= 0
                    else station[1] - self.cl
                    for station in self.stations]
        else:
            slon = [station[1] for station in self.stations]

        ax = plt.gca()
        ax.scatter(slon, slat, s=20, marker='v', c=((0.85, 0.2, 0.2),),
                   edgecolors='k', linewidths=0.25, zorder=-1)

    def plot_histogram(self, ddata, n_bins, facecolor=(0.8, 0.8, 0.8),
                       alpha=1):
        """Plots histogram of input data."""

        # the histogram of the data
        ax = plt.gca()
        n, bins, patches = ax.hist(ddata, n_bins, facecolor=facecolor,
                                   alpha=alpha)

    def plot_xcorr_matrix(self):
        """Plots Corrlation matrix with approximate correlation bars
        """

        fig = plt.figure(figsize=(12, 11))

        ax = fig.subplots(10, 10, sharex="col", sharey='row', squeeze=True,
                          gridspec_kw={'hspace': 0, 'wspace': 0})

        for _i in range(10):
            for _j in range(10):

                plt.sca(ax[_i][_j])
                if _j == _i:
                    shay = ax[_i][_j].get_shared_y_axes()
                    shay.remove(ax[_i][_j])
                    self.plot_histogram(self.dCMT[:, _i], self.nbins)
                else:
                    ax[_i][_j].plot(self.dCMT[:, _j],  self.dCMT[:, _i],
                                    'ko', markersize=2)

                    # OLS fit
                    A = np.vstack([self.dCMT[:, _j],
                                   np.ones(len(self.dCMT[:, _j]))]).T

                    m, c = np.linalg.lstsq(A, self.dCMT[:, _i], rcond=None)[0]

                    res = np.sqrt(np.sum(((c + m * self.dCMT[:, _j])
                                          - self.dCMT[:, _j]) ** 2))

                    print(res, np.sqrt(self.mean_mat[_j]**2
                                       + self.mean_mat[_i]**2))

                    if res < 0.25 * self.N*np.sqrt(self.mean_mat[_j]**2
                                                   + self.mean_mat[_i]**2):
                        # Different option to compute the OLS fit by computing
                        # the perpendicular distance
                        # m, c = fit_xy(self.dCMT[:, _j], self.dCMT[:, _i])

                        # Plot polyline
                        ax[_i][_j].plot(self.dCMT[:, _j],
                                        c + m * self.dCMT[:, _j],
                                        '-', c=(0.85, 0.2, 0.2))

        plt.tight_layout()
        for _i in range(10):
            for _j in range(10):
                if _i == 9:
                    ax[_i][_j].set_xlabel(self.dlabels[_j])
                if _j == 0:
                    ax[_i][_j].set_ylabel(self.dlabels[_i])
                    ax[_i][_j].yaxis.set_label_coords(-0.3, 0.5)

                    # Label magic happens here
                    if _i in [0, 1, 2, 3, 4, 5, 6]:
                        # Change the ticklabel format to scientific format
                        ax[_i][_j].ticklabel_format(axis="y", style='sci')
                        offset_text = \
                            ax[_i][_j].yaxis.get_offset_text().get_text()
                        ax[_i][_j].text(-.375, 0.85, offset_text,
                                        rotation='vertical',
                                        ha='center', va='center',
                                        transform=ax[_i][_j].transAxes)
                        ax[_i][_j].yaxis.get_offset_text().set_visible(False)

        # Finally plot shot
        plt.tight_layout()
        if self.savedir is not None:
            plt.savefig(os.path.join(self.savedir, "xcorr.pdf"))
        else:
            plt.show()

    def plot_measurement_changes(self):
        """ Uses the stat_dict to compute measurement statistics.

        """

        # Create figure
        if self.stat_dict is None:
            logger.info("No statistics dictionary here...")
            return

        # Get number of rows and columns
        nrows = len(self.stat_dict)
        ncols = len(self.vtype_dict)

        stats = ["mean", "std"]
        # Create figure

        for _stat in stats:

            fig = plt.figure(figsize=(4 * ncols, 4 * nrows))
            G = GridSpec(nrows, ncols)

            for irow, cat in enumerate(self.stat_dict.keys()):
                for icol, (label, vtype) in enumerate(self.vtype_dict.items()):
                    ax = fig.add_subplot(G[irow, icol])
                    if irow == 0 and icol == 0:
                        fontsize = 20
                        ax.text(0.025, 0.975, _stat,
                                fontsize=fontsize,
                                horizontalalignment='left',
                                verticalalignment='top',
                                transform=ax.transAxes)

                    self._hist_sub(self.stat_dict[cat][vtype], cat, vtype,
                                   label, _stat, self.nbins)

            plt.tight_layout()
            plt.savefig(os.path.join(self.savedir,
                                     "measurement_changes_" + _stat + ".pdf"))
            plt.close(fig)

    def plot_mean_measurement_change_stats(self):
        """ Uses the stat_dict to compute measurement statistics.

        """

        # Create figure
        if self.stat_dict is None:
            logger.info("No statistics dictionary here...")
            return

        # Get number of rows and columns
        nrows = len(self.stat_dict)
        ncols = len(self.vtype_dict)

        # Create figure
        fig = plt.figure(figsize=(4 * ncols, 4 * nrows))
        G = GridSpec(nrows, ncols)

        for irow, cat in enumerate(self.stat_dict.keys()):
            for icol, (label, vtype) in enumerate(self.vtype_dict.items()):
                ax = fig.add_subplot(G[irow, icol])
                if irow == 0 and icol == 0:
                    fontsize = 20
                    ax.text(0.025, 0.975, r"$\Delta$Mean",
                            fontsize=fontsize,
                            horizontalalignment='left',
                            verticalalignment='top',
                            transform=ax.transAxes)

                self._hist_sub_change(self.stat_dict[cat][vtype], cat, vtype,
                                      label, self.nbins)

        plt.tight_layout()
        plt.savefig(os.path.join(self.savedir,
                                 "mean_measurement_change_stats.pdf"))
        plt.close(fig)

    @staticmethod
    def _hist_sub(measurement_dict, cat, vtype, label, stat, num_bin):
        """

        :param measurement_dict: {after: {mean: list, std: list},
                                  before: {mean: list, std: list}}
        :param label:
        :param stat:
        :return:
        """
        # Get axes
        ax = plt.gca()

        plt.xlabel(label, fontsize=15)
        plt.ylabel(r"%s" % cat.replace("_", r"\_"), fontsize=15)

        data_b = measurement_dict['before'][stat]
        data_a = measurement_dict['after'][stat]

        if vtype == "cc":
            ax_min = min(min(data_b), min(data_a))
            ax_max = max(max(data_b), max(data_a))
        elif vtype == "chi":
            ax_min = 0.0
            ax_max = max(max(data_b), max(data_a))
        else:
            ax_min = min(min(data_b), min(data_a))
            ax_max = max(max(data_b), max(data_a))
            abs_max = max(abs(ax_min), abs(ax_max))
            ax_min = -abs_max
            ax_max = abs_max

        if stat == "std":
            ax_min = 0.0

        binwidth = (ax_max - ax_min) / num_bin

        # Stats
        a_mean = np.mean(data_a)
        a_std = np.std(data_a)
        b_mean = np.mean(data_b)
        b_std = np.std(data_b)

        nb, _, _ = ax.hist(
            data_b, bins=np.arange(ax_min, ax_max + binwidth / 2., binwidth),
            facecolor='blue', alpha=0.3)
        nb_max = np.max(nb)
        ax.plot([b_mean, b_mean], [0, nb_max], "b--")
        ax.plot([b_mean - b_std, b_mean + b_std],
                [nb_max / 2, nb_max / 2], "b--")
        na, _, _ = ax.hist(
            data_a, bins=np.arange(ax_min, ax_max + binwidth / 2., binwidth),
            facecolor='red', alpha=0.5)
        na_max = np.max(na)
        ax.plot([a_mean, a_mean], [0, na_max], "r-")
        ax.plot([a_mean - a_std, a_mean + a_std],
                [na_max / 2, na_max / 2], "r-")

    @staticmethod
    def _hist_sub_change(measurement_dict, cat, vtype, label, num_bin):
        """

        :param measurement_dict: {after: {mean: list, std: list},
                                  before: {mean: list, std: list}}
        :param label:
        :param stat:
        :return:
        """
        # Get axes
        ax = plt.gca()

        plt.xlabel(label, fontsize=15)
        plt.ylabel(r"%s" % cat.replace("_", r"\_"), fontsize=15)

        # Compute data change
        data_b = measurement_dict['before']["mean"]
        data_a = measurement_dict['after']["mean"]
        data = (np.array(data_a) - np.array(data_b)).tolist()

        # Compute plot axes and bins
        ax_min = min(data)
        ax_max = max(data)
        abs_max = max(abs(ax_min), abs(ax_max))
        ax_min = -abs_max
        ax_max = abs_max
        binwidth = (ax_max - ax_min) / num_bin

        # Stats
        d_mean = np.mean(data)
        d_std = np.std(data)

        nd, _, _ = ax.hist(
            data, bins=np.arange(ax_min, ax_max + binwidth / 2., binwidth),
            facecolor='black', alpha=0.3)
        nd_max = np.max(nd)
        ax.plot([d_mean, d_mean], [0, nd_max], "b-")
        ax.plot([d_mean - d_std, d_mean + d_std],
                [nd_max / 2, nd_max / 2], "b-")

    def plot_xcorr_heat(self):
        """Plots correlatio n heatmap.
        """
        fig = plt.figure(figsize=(12, 10))

        ax = fig.add_subplot(111)

        x = np.arange(10)
        y = np.arange(10)
        xx, yy = np.meshgrid(x, y)

        cmap = plt.get_cmap("coolwarm")
        size_scale = 750
        scat = ax.scatter(xx, yy, s=(np.abs(self.xcorr_mat[:10, :10])
                                     * size_scale),
                          c=self.xcorr_mat[:10, :10], cmap=cmap, marker='s',
                          alpha=0.75)

        cbar = plt.colorbar(scat, aspect=50, pad=0)
        cbar.ax.tick_params(labelsize=12)

        # Mapping from column names to integer coordinates
        plt.xticks(np.arange(10), self.dlabels[:10], fontsize=13)
        plt.yticks(np.arange(10), self.dlabels[:10], fontsize=13)

        # Relocate Grid
        ax.grid(False, 'major')
        ax.grid(True, 'minor')
        ax.set_xticks([t + 0.5 for t in ax.get_xticks()], minor=True)
        ax.set_yticks([t + 0.5 for t in ax.get_yticks()], minor=True)

        # Set limits
        ax.set_xlim([-0.5, max(x) + 0.5])
        ax.set_ylim([-0.5, max(x) + 0.5])

        # Invert y and put xaxis on top
        ax.invert_yaxis()
        ax.xaxis.tick_top()

        # Finally plot shot
        plt.tight_layout()
        if self.savedir is not None:
            plt.savefig(os.path.join(self.savedir, "xcorrheat.pdf"))
        else:
            plt.show()

    def plot_table(self):
        """Plots minimal summary"""

        columns = (r'$\mu$', r'$\sigma$')
        rows = [r'$\delta t$', r'$\delta$Lat', r'$\delta$Lon', r'$\delta z$',
                r'$\delta M_0$', r"$\delta M_{rr}$", r"$\delta M_{tt}$",
                r"$\delta M_{pp}$", r"$\delta M_{rt}$", r"$\delta M_{rp}$",
                r"$\delta M_{tp}$"]

        cell_text = []

        # dt
        cell_text.append(["%3.3f" % (self.mean_mat[10]),
                          "%3.3f" % (self.std_mat[10])])
        # dLat
        cell_text.append(["%3.3f" % (self.mean_mat[8]),
                          "%3.3f" % (self.std_mat[8])])
        # dLon
        cell_text.append(["%3.3f" % (self.mean_mat[9]),
                          "%3.3f" % (self.std_mat[9])])
        # dz
        cell_text.append(["%3.3f" % (self.mean_mat[7]),
                          "%3.3f" % (self.std_mat[7])])
        # M0
        cell_text.append(["%3.3f" % (self.mean_mat[0]),
                          "%3.3f" % (self.std_mat[0])])
        for _j in range(6):
            cell_text.append(["%3.3f" % (self.mean_mat[1 + _j]),
                              "%3.3f" % (self.std_mat[1 + _j])])

        # Plot table
        ax = plt.gca()
        ax.axis('tight')
        ax.axis('off')
        ax.table(cellText=cell_text, rowLabels=rows, colLabels=columns,
                 loc='center', edges='horizontal', fontsize=13)

    def save_table(self):
        """Uses plot_table function to save figure"""

        # Create figure handle
        fig = plt.figure(figsize=(4, 2))

        # Create subplot layout
        self.plot_table()

        plt.savefig(os.path.join(self.savedir, "summary_table.pdf"))
        plt.close(fig)


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)
"""

from __future__ import (absolute_import, division, print_function)

from .window import window_on_stream, window_on_trace  # NOQA



"""
This file contains functions and classes to check whether the source
inversion works

:copyright:
    Lucas Sawade (lsawade@princeton.edu, 2019)

:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)

"""

from . import forward as fw
from .plot import PlotSTFInversion
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
matplotlib.rcParams['mathtext.fontset'] = 'cm'
matplotlib.rcParams['font.family'] = 'STIXGeneral'
matplotlib.rcParams['font.weight'] = 'bold'
matplotlib.rcParams["axes.labelweight"] = "bold"
matplotlib.rcParams['text.usetex'] = True


def main():

    # Start at random seed
    np.random.seed(260789)

    # Create synthetic data
    nt = 3002
    dt = 0.05
    nr = 20
    dx = 10.

    green_vel = np.array([20,  19,  19, 10, 13])

    # Data for Vertical component Z
    # delta_loc_tz = np.array([50, 55, 57, 78, 96])
    delta_loc_tz = np.array([25, 40, 57, 78, 96])

    delta_ampz = 3*dt*np.array([4,  -2,  -1, 3, 0.5])
    t, xrz, Gz, synz, obsz, wavelet = \
        fw.synthetic_traces(nr=nr, dx=dx, nt=nt, dt=0.05,
                            green_vel=green_vel,
                            delta_loc_t=delta_loc_tz,
                            delta_amp=delta_ampz,
                            amp1=2.5, to1=20, sig1=5,
                            amp2=1.75, to2=32, sig2=3)

    # # Data for radial R
    # delta_loc_tr = np.array([45, 53, 55, 75, 95])
    # delta_ampr = 3 * dt * np.array([4, -2.5, -.5, 3.5, -.5])
    # t, xrr, Gr, synr, obsr, waveletr = \
    #     fw.synthetic_traces(nr=nr, dx=dx, nt=nt, dt=0.05,
    #                         green_vel=green_vel,
    #                         delta_loc_t=delta_loc_tr,
    #                         delta_amp=delta_ampr,
    #                         amp1=2.5, to1=20, sig1=5,
    #                         amp2=1.75, to2=32, sig2=3)
    #
    # # Data for transverse T
    # delta_loc_tt = np.array([48, 53, 60, 75, 97])
    # delta_ampt = 3 * dt * np.array([3.5, -1.5, -2.5, 1, 1.5])
    # t, xrt, Gt, synt, obst, wavelet = \
    #     fw.synthetic_traces(nr=nr, dx=dx, nt=nt, dt=0.05,
    #                         green_vel=green_vel,
    #                         delta_loc_t=delta_loc_tt,
    #                         delta_amp=delta_ampt,
    #                         amp1=2.5, to1=20, sig1=5,
    #                         amp2=1.75, to2=32, sig2=3)

    # Create full matrices
    # G = np.concatenate((Gr, Gt, Gz), axis=0)
    # syn = np.concatenate((synr, synt, synz), axis=0)
    # obs = np.concatenate((obsr, obst, obsz), axis=0)

    G = Gz
    syn = synz
    obs = obsz

    # # Plot one trace
    # fw.plot_one_trace_set(t, wavelet, G[0, :], syn[0, :], obs[0, :])

    # # Plot section
    # fw.plot_section(t, syn, obs)

    # Plot one trace
    fw.plot_one_trace_set(t, wavelet, G[0, :], syn[0, :], obs[0, :])

    # Plot section
    fw.plot_section(t, syn, obs)

    # Compute Source time function with conventional waterlevel deconvolution
    lambd = 0.01  # waterlevel
    stf_decon, syn_decon1 = fw.deconvolution(obs, G, lambd)

    # Compute source time function with iterative landweber
    stf_lw, stf_list, chi_list = fw.landweber(obs, G, dt, maxT=50, crit=0.01,
                                              lamb=0.001, type="2")

    # # Compare Source Time functions
    # fw.plot_source_comparison(t, wavelet, [stf_decon, stf_lw],
    #                           ['Estimate LS', 'Landweber'])

    # Compute new synthetics
    syn_decon = fw.compute_synth(G, stf_decon)
    syn_lw = fw.compute_synth(G, stf_lw)

    # # Plot section comparison
    # fw.plot_comparison_section(t, obs, syn, syn_decon, syn_lw)

    # # Plot Stf evolution
    # fw.plot_stf_evolution(t, wavelet, stf_list, skip=5)

    PL = PlotSTFInversion(t=t, stf=wavelet, G=G, obs=obs, syn=syn,
                          syn_decon=syn_decon, syn_lw=syn_lw,
                          stf_decon=stf_decon, stf_lw=stf_lw,
                          stf_list=stf_list, skip=5, save_dir="./")

    PL.plot_result()

    # Plot section comparison
    fw.plot_comparison_section(t, obs, syn, syn_decon, syn_lw)

    # Plot Stf evolution
    fw.plot_stf_evolution(t, wavelet, stf_list, skip=5)

    # Compute
    plt.show()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This script will download the observed data. To the necessary places.


:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""

from gcmt3d.workflow.data_request3 import data_request
import argparse


def main():
    # Create argument parser
    parser = argparse.ArgumentParser()

    # Set arguments
    parser.add_argument('-f', dest='filename', required=True,
                        help='Path to CMTSOLUTION file in database',
                        type=str)

    parser.add_argument('-p', dest='param_path', required=True,
                        help='Path to parameter directory',
                        type=str)

    # Get Arguments
    args = parser.parse_args()

    data_request(args.filename, args.param_path)


if __name__ == "__main__":
    main()



"""

This script contains functions to create the skeleton structure for the GCMT3D
database of Earthquakes. If wanted.

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.orcopyleft/gpl.html)

Last Update: June 2019

"""

import glob
import os
import shutil
import warnings
from distutils.dir_util import copy_tree
from obspy import read_events
from time import time
from ...source import CMTSource
from ...asdf.utils import write_yaml_file

import logging
from ...log_util import modify_logger

# Create logger
logger = logging.getLogger(__name__)
modify_logger(logger)

STATIONS = os.path.join(os.path.dirname(os.path.abspath(__file__)), "STATIONS")


class DataBaseSkeleton(object):
    """Class to handle data skeleton creation. If specfem directory is given,
    the class copies the necessary data from the specfem folder."""

    def __init__(self, basedir=None, cmt_fn=None, specfem_dir=None,
                 stations_file=None, npar=9,
                 overwrite=False):
        """
        Args:
            basedir: str with path to database directory, e.g
                            ".../path/to/database/"
            cmt_fn: path to cmt solution, e.g ".../path/to/CMTSOLUTION", or,
                            e.g ".../path/to/cmtsolutionfile with list."
            npar: number of parameters to be inverted for. 6,7 or 9. Default
                  9
            specfem_dir: str with path to specfem directory, e.g
                            ".../path/to/specfem3d_globe"
            overwrite: number which sets what should be over written. If not
                       a number or False, nothing will be overwritten.
                        0: Everything is overwritten, including the database
                        1: The specific earthquake is overwritten:
                        2: Seismogram, Simulation and Window Directories are
                           overwritten for a certain Earthquake.
                        3: Simulation subdirectories overwritten
                        4: Overwrite yaml path files only
            verbose: boolean that sets whether the output should be verbose
                     or not

        """

        self.basedir = basedir
        self.specfem_dir = specfem_dir
        self.npar = npar

        # Modifiers
        self.ow = overwrite

        # Check if things exists
        self.cmt_fn = cmt_fn

        # Check if cmt file name exists and if there are more than one
        self.cmtfile_list = glob.glob(self.cmt_fn)

        # Throw error if list empty
        if self.cmtfile_list is []:
            raise ValueError("No CMTSOLUTION file exists of that name.")

        # Create empty earthquake directory list
        self.Cdirs = []
        self.Cids = []

        # List of possible attributes
        # Parameters
        self.attr = ["CMT", "CMT_rr", "CMT_tt", "CMT_pp", "CMT_rt", "CMT_rp",
                     "CMT_tp", "CMT_depth", "CMT_lat", "CMT_lon"]

        if stations_file is None:
            self.stations_file = STATIONS
        else:
            self.stations_file = stations_file

    def create_all(self):
        """ Writes complete database structure."""

        logger.info("Creating earthquake entry in database %s ... "
                    % self.basedir)
        t1 = time()

        # Create earthquake directory
        logger.info("Creating simulation directories...")
        self.create_Cdirs()

        # Create station metadata directory
        logger.info("Creating station directory...")
        self.create_station_dir()

        # Create Log directory
        logger.info("Creating log directory...")
        self.create_log_dir()

        # Create window data directory
        logger.info("Creating window...")
        self.create_window_dir()

        # Create Seismogram directory
        logger.info("Creating seismogram directories")
        self.create_seismogram_dir()

        # Create Inversion directory
        logger.info("Creating inversion directories")
        self.create_inversion_dir()

        if self.specfem_dir is not None:
            # Create
            self.create_CMT_SIM_dir()

        logger.info("Done. Time Elapsed %.1f s" % (time() - t1))

    def create_base(self):
        """Creates Base directory if it doesn't exist."""

        # Overwrite 0 overwrite total database
        if self.ow in [0] and type(self.ow) is not bool:
            self._create_dir(self.basedir, True)
        else:
            self._create_dir(self.basedir, False)

    def create_Cdirs(self):
        """ If more than one earthquake exist with regex, call all of them"""
        for cmtfile in self.cmtfile_list:

            # Create directory
            self.create_1_Cdir(cmtfile)

    def create_1_Cdir(self, cmtfile):
        """Creates 1 Earthquake directory"""

        # create CMT
        cmt = CMTSource.from_CMTSOLUTION_file(cmtfile)

        # Create CMTSource to extract the file name
        Cid = cmt.eventname
        self.Cids.append(Cid)

        # Earthquake directory
        Cdir = os.path.join(self.basedir, "C" + Cid)

        # Create directory
        if self.ow in [0, 1] and type(self.ow) is not bool:
            self._create_dir(Cdir, True)
        else:
            self._create_dir(Cdir, False)

        # Append directory path to the list.
        self.Cdirs.append(Cdir)

        # Create new CMT path
        cmt_path = os.path.join(Cdir, "C" + Cid + ".cmt")

        # Create new CMT path
        xml_path = os.path.join(Cdir, "C" + Cid + ".xml")

        # Copy the Earthquake file into the directory with C<ID>.cmt
        if self.ow in [0, 1] and type(self.ow) is not bool:
            self._copy_file(cmtfile, cmt_path, True)
        else:
            self._copy_file(cmtfile, cmt_path, False)

        # Copy the Earthquake file into the directory with C<ID>.xml
        if self.ow in [0, 1] and type(self.ow) is not bool:
            self._write_quakeml(cmtfile, xml_path, True)
        else:
            self._write_quakeml(cmtfile, xml_path, False)

    def create_station_dir(self):
        """Creates station_data directory for station metadata."""

        for _i, _Cdir in enumerate(self.Cdirs):

            # Create station_data dirs
            station_dir = os.path.join(_Cdir, "station_data")

            if self.ow in [0, 1, 2] and type(self.ow) is not bool:
                # Create new directory
                self._create_dir(station_dir, True)
            else:
                self._create_dir(station_dir, False)

            # Set new destination path
            dst_path0 = os.path.join(station_dir, "STATIONS")

            if self.ow in [0, 1, 2, 3] \
                    and type(self.ow) is not bool:
                self._copy_file(self.stations_file, dst_path0, True)
            else:
                self._copy_file(self.stations_file, dst_path0, False)

    def create_log_dir(self):
        """Creates station_data directory for station metadata."""

        for _i, _Cdir in enumerate(self.Cdirs):

            # Create station_data dirs
            log_dir = os.path.join(_Cdir, "logs")

            if self.ow in [0, 1, 2] and type(self.ow) is not bool:
                # Create new directory
                self._create_dir(log_dir, True)
            else:
                self._create_dir(log_dir, False)

    def create_inversion_dir(self):
        """Creates station_data directory for station metadata."""

        for _i, _Cdir in enumerate(self.Cdirs):

            # Create station_data dirs
            inv_dir = os.path.join(_Cdir, "inversion")

            if self.ow in [0, 1, 2] and type(self.ow) is not bool:
                # Create new directory
                self._create_dir(inv_dir, True)
            else:
                self._create_dir(inv_dir, False)

            inv_dict_dir = os.path.join(inv_dir, "inversion_dicts")

            if self.ow in [0, 1, 2, 3] and type(self.ow) is not bool:
                # Create new directory
                self._create_dir(inv_dict_dir, True)
            else:
                self._create_dir(inv_dict_dir, False)

            inv_out_dir = os.path.join(inv_dir, "inversion_output")

            if self.ow in [0, 1, 2, 3] and type(self.ow) is not bool:
                # Create new directory
                self._create_dir(inv_out_dir, True)
            else:
                self._create_dir(inv_out_dir, False)

            inv_types = ["cmt3d", "g3d"]

            for _inv_type in inv_types:

                type_dir = os.path.join(inv_out_dir, _inv_type)

                if self.ow in [0, 1, 2, 3, 4] and type(self.ow) is not bool:
                    # Create new directory
                    self._create_dir(type_dir, True)
                else:
                    self._create_dir(type_dir, False)

                inv_new_syn_dir = os.path.join(type_dir, "new_synt")

                if self.ow in [0, 1, 2, 3, 4] and type(self.ow) is not bool:
                    # Create new directory
                    self._create_dir(inv_new_syn_dir, True)
                else:
                    self._create_dir(inv_new_syn_dir, False)

                inv_new_syn_plots_dir = os.path.join(type_dir,
                                                     "waveform_plots")

                if self.ow in [0, 1, 2, 3, 4] and type(self.ow) is not bool:
                    # Create new directory
                    self._create_dir(inv_new_syn_plots_dir, True)
                else:
                    self._create_dir(inv_new_syn_plots_dir, False)

    def create_window_dir(self):
        """Creates window_data directory for pyflex window data metadata."""

        for _i, _Cdir in enumerate(self.Cdirs):

            # Create window_data dirs
            window_dir = os.path.join(_Cdir, "window_data")

            if self.ow in [0, 1, 2] and type(self.ow) is not bool:
                # Create new directory
                self._create_dir(window_dir, True)
            else:
                self._create_dir(window_dir, False)

            # Create Window path directory
            window_path_dir = os.path.join(window_dir, "window_paths")

            if self.ow in [0, 1, 2, 3] and type(self.ow) is not bool:
                # Create new directory
                self._create_dir(window_path_dir, True)
            else:
                self._create_dir(window_path_dir, False)

    def create_CMT_SIM_dir(self):
        """
        Creates CMT simulation directory and copies necessary files from given
        specfem directory. Important here is the fact that specfem has to be
        compiled already and there is no way of testing that prior to running
        GCMT.
        """

        for _i, _eq in enumerate(self.Cdirs):

            # First create main directory
            sim_path = os.path.join(_eq, "CMT_SIMs")

            if self.ow in [0, 1, 2] and type(self.ow) is not bool:
                self._create_dir(sim_path, True)
            else:
                self._create_dir(sim_path, False)

            # Second create subdirectories of CMT specfem directories
            for _j, _attr in enumerate(self.attr[:self.npar+1]):

                # Create subdirectory for simulation packages.
                cmt_der_path = os.path.join(sim_path, _attr)

                if self.ow in [0, 1, 2, 3] and type(self.ow) is not bool:
                    self._create_dir(cmt_der_path, True)
                else:
                    self._create_dir(cmt_der_path, False)

                # Copy specfem directory into cmt_der_path
                subdirs = ["DATA", "OUTPUT_FILES"]

                for _k, _subdir in enumerate(subdirs):
                    # Path to specfem subdirectory
                    src_path = os.path.join(self.specfem_dir, _subdir)

                    # Path to destination directory
                    dst_path = os.path.join(cmt_der_path, _subdir)

                    # Only copy some files from the DATA directory
                    if _subdir == "DATA":

                        # Create DATA directory
                        if self.ow in [0, 1, 2, 3] and type(
                                self.ow) is not bool:
                            self._create_dir(dst_path, True)
                        else:
                            self._create_dir(dst_path, False)

                        # Set files to be copied into directory
                        files = ["Par_file", "CMTSOLUTION"]

                        for file in files:

                            # Set new source path
                            src_path0 = os.path.join(src_path, file)

                            # Set new destination path
                            dst_path0 = os.path.join(dst_path, file)

                            if self.ow in [0, 1, 2, 3] and type(
                                    self.ow) is not bool:
                                self._copy_file(src_path0, dst_path0, True)
                            else:
                                self._copy_file(src_path0, dst_path0, False)

                        # Set new destination path
                        dst_path0 = os.path.join(dst_path, "STATIONS")

                        if self.ow in [0, 1, 2, 3] \
                                and type(self.ow) is not bool:
                            self._copy_file(self.stations_file,
                                            dst_path0, True)
                        else:
                            self._copy_file(self.stations_file,
                                            dst_path0, False)

                    else:
                        if self.ow in [0, 1, 2, 3] and type(
                                self.ow) is not bool:
                            self._copy_dir(src_path, dst_path, True)
                        else:
                            self._copy_dir(src_path, dst_path, False)

                    # Create the Path file for later usage of the ASDF
                    # conversion
                    if _subdir == "OUTPUT_FILES":
                        logger.verbose("Writing YAML path file "
                                       "from waveform dir "
                                       "%s" % dst_path)
                        self._create_syn_path_yaml(dst_path)

                # Create symbolic link to destination folders
                if not os.path.islink((os.path.join(cmt_der_path, "bin"))):
                    os.symlink(os.path.join(self.specfem_dir, "bin"),
                               os.path.join(cmt_der_path, "bin"),
                               target_is_directory=True)

                # Create symbolic link to destination folders
                if not os.path.islink((os.path.join(cmt_der_path,
                                                    "DATABASES_MPI"))):
                    os.symlink(os.path.join(self.specfem_dir,
                                            "DATABASES_MPI"),
                               os.path.join(cmt_der_path,
                                            "DATABASES_MPI"),
                               target_is_directory=True)

    def create_seismogram_dir(self):
        """Creates response subdirectory"""

        for _i, _Cdir in enumerate(self.Cdirs):
            # Create response path
            seismogram_dir = os.path.join(_Cdir, "seismograms")

            if self.ow in [0, 1, 2] and type(self.ow) is not bool:
                # Create new directory
                self._create_dir(seismogram_dir, True)
            else:
                self._create_dir(seismogram_dir, False)

            # Create the subdirectory for synthetic data
            syn_dir_path = os.path.join(seismogram_dir, 'syn')

            if self.ow in [0, 1, 2] and type(self.ow) is not bool:
                # Create new directory
                self._create_dir(syn_dir_path, True)
            else:
                self._create_dir(syn_dir_path, False)

            # Create the subdirectory for observed data
            obs_dir_path = os.path.join(seismogram_dir, 'obs')

            if self.ow in [0, 1, 2] and type(self.ow) is not bool:
                # Create new directory
                self._create_dir(obs_dir_path, True)
            else:
                self._create_dir(obs_dir_path, False)

            # write Observed path file
            logger.verbose("Writing the YAML path file to %s" %
                           os.path.join(obs_dir_path, "observed.yml"))

            self._create_obs_path_yaml(self.Cids[_i], _Cdir)

            # Create the subdirectory for processing pathfiles
            process_dir_path = os.path.join(seismogram_dir, 'process_paths')

            if self.ow in [0, 1, 2] and type(self.ow) is not bool:
                # Create new directory
                self._create_dir(process_dir_path, True)
            else:
                self._create_dir(process_dir_path, False)

            # Create the subdirectory for processed seismograms
            processed_dir_path = os.path.join(seismogram_dir,
                                              'processed_seismograms')

            if self.ow in [0, 1, 2] and type(self.ow) is not bool:
                # Create new directory
                self._create_dir(processed_dir_path, True)
            else:
                self._create_dir(processed_dir_path, False)

    def _copy_dir(self, source, destination, ow, **kwargs):
        """ Copies a directory source to destination. It checks also for
        potential duplicates in the same place."""

        if os.path.isdir(destination) and ow:
            logger.verbose("Directory %s exists already. It will "
                           "be overwritten." % destination)
            shutil.rmtree(destination)
            copy_tree(source, destination, **kwargs)

        elif os.path.isdir(destination) and ow is False:
            logger.verbose("Directory %s exists already. It will "
                           "NOT be overwritten." % destination)
        else:
            logger.verbose("Copying directory %s file to %s"
                           % (source, destination))
            copy_tree(source, destination)

    def _copy_file(self, source, destination, ow):
        """ Copies file from source to destination. """

        if os.path.isfile(destination) and ow:
            logger.verbose("File %s exists already. It will "
                           "be overwritten." % destination)
            self._replace_file(source, destination)

        elif os.path.isfile(destination) and ow is False:
            logger.verbose("File %s exists already. It will "
                           "NOT be overwritten." % destination)
        else:
            logger.verbose("Copying file %s file to %s."
                           % (source, destination))
            shutil.copyfile(source, destination)

    def _write_quakeml(self, source, destination, ow):
        """ Copies CMTSOLUTION from source to QuakeML destination. It checks
        also for potential duplicates in the same place, warns whether they are
        different but have the name."""

        # CMT Source file
        catalog = read_events(source)

        if os.path.isfile(destination) and ow:
            logger.verbose("Earthquake file %s exists already. It will "
                           "be overwritten." % destination)
            os.remove(destination)
            catalog.write(destination, format="QUAKEML")

        elif os.path.isfile(destination) and ow is False:
            logger.verbose("Earthquake file %s exists already. It will "
                           "NOT be overwritten." % destination)

            # Warn if destination eq is not equal to new eq
            if not CMTSource.from_CMTSOLUTION_file(source) \
                    == CMTSource.from_quakeml_file(destination):
                warnings.warn("CMT solution in the database is not "
                              "the same as the file with the same ID.")
        else:
            logger.verbose("Writing earthquake %s file to %s." % (source,
                                                                  destination))
            catalog.write(destination, format="QUAKEML")

    def _create_syn_path_yaml(self, waveform_dir):
        """ This function writes a yaml conversion path file for 1 Simulation
        file. This file is later on need for the creation of ASDF files and the
        processing involved ASDF files.

        The function assumes that
        * the QuakeML file is located in the OUTPUT_FILES directory with the
          name `Quake.xml`
        * The output directory name is the
          `../database/C<id>/seismograms/syn/<attr>.h5

        Args:
              waveform_dir: path to OUTPUT_FILES

        """

        # File Type
        filetype = "sac"

        # Tag
        tag = "syn"

        # QuakeML file path
        quakeml_file = os.path.join(waveform_dir, "Quake.xml")

        # Outputfile
        cmt_sim_dir = os.path.dirname(waveform_dir)
        Cdir = os.path.dirname(os.path.dirname(cmt_sim_dir))
        cmt_name = os.path.basename(cmt_sim_dir)
        output_file = os.path.join(Cdir, "seismograms", "syn",
                                   cmt_name + ".h5")

        # Pathfile directory
        yaml_file_path = os.path.join(cmt_sim_dir, cmt_name + ".yml")

        # Create dictionary
        logger.verbose("Writing path file %s." % yaml_file_path)

        d = {"waveform_dir": waveform_dir,
             "filetype": filetype,
             "quakeml_file": quakeml_file,
             "tag": tag,
             "output_file": output_file}

        # Writing the directory to file
        write_yaml_file(d, yaml_file_path)

    def _create_obs_path_yaml(self, Cid, Cdir):
        """ This function writes a yaml path file for 1 Simulation file. This
        file is later on need for the creation of ASDF files and the
        processing involved ASDF files.

        The function assumes that
        * the QuakeML file is located in the main EQ directory with the
          name `C<id>.xml`
        * The output file name is the
          `../database/C<id>/seismograms/obs/raw_observed.h5

        Args:
              waveform_dir: path to OUTPUT_FILES

        """

        # Tag
        tag = "obs"

        # Waveform file
        waveform_files = os.path.join(Cdir, "seismograms", "obs",
                                      "*.mseed")

        # QuakeML file path
        quakeml_file = os.path.join(Cdir, "C" + Cid + ".xml")

        # Station file
        staxml_file = os.path.join(Cdir, "station_data", "*.xml")

        # Outputfile
        output_file = os.path.join(Cdir, "seismograms",
                                   "obs", "raw_observed.h5")

        # Pathfile directory
        yaml_file_path = os.path.join(Cdir, "seismograms",
                                      "obs", "observed.yml")

        # Create dictionary
        logger.verbose("Writing path file %s." % yaml_file_path)

        d = {"waveform_files": waveform_files,
             "quakeml_file": quakeml_file,
             "tag": tag,
             "staxml_files": staxml_file,
             "output_file": output_file}

        # Writing the directory to file
        write_yaml_file(d, yaml_file_path)

    def _create_dir(self, directory, ow):
        """Create subdirectory"""
        if os.path.exists(directory) and os.path.isdir(directory) \
                and ow is False:
            logger.verbose("%s exists already. Not overwritten." % directory)

        elif os.path.exists(directory) and os.path.isdir(directory) \
                and ow is True:
            logger.verbose("%s exists already, but overwritten." % directory)
            self._replace_dir(directory)
        else:
            logger.verbose("Creating directory %s." % directory)
            os.makedirs(directory)

    @staticmethod
    def _replace_dir(destination, source=None):
        """Mini function that replaces a directory"""
        if os.path.exists(destination) and os.path.isdir(destination):
            shutil.rmtree(destination)
        if source is None:
            os.makedirs(destination)

    @staticmethod
    def _replace_file(source, destination):
        """Mini function that replaces a file"""
        if os.path.exists(destination) and os.path.isfile(destination):
            os.remove(destination)
        shutil.copyfile(source, destination)

    def __str__(self):
        """This is what is printed when the initialized class is called in the
        print function. the function returns a string that contains the paths
        to the different directories specified."""

        string = "Database directory:\n"
        string += self.basedir + "\n\n"
        string += "CMTSOLUTION path:\n"
        string += self.cmt_fn + "\n\n"
        string += "Specfem directory:\n"
        string += self.specfem_dir + "\n\n"
        return string



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Filter the window based on the sensor type. For example, in long
period band(90-250s), we want to keep only STS-1 instrument windows.
For the input file, it requires 1) sensor type as json file; 2) windows
as json file. For the output, it is going to replace the origin window file
and keep a copy of original windows as "***.origin.json"

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""
from __future__ import (absolute_import, division, print_function)
import os
import argparse
from pprint import pprint
from gcmt3d.window.filter_windows import filter_windows, count_windows
from gcmt3d.asdf.utils import load_json, dump_json, load_yaml


def check_keys(dictv, keys):
    set1 = set(dictv.keys())
    set2 = set(keys)
    if set1 != set2:
        print("Missing keys: %s" % (set2 - set1))
        print("Redundant keys: %s" % (set1 - set2))
        return False
    return True


def check_path(paths):
    keys = ["window_file", "station_file", "output_file", "measurement_file"]
    if not check_keys(paths, keys):
        raise ValueError("Path file is bad!")

    print("=" * 10 + " Path info " + "=" * 10)
    pprint(paths)


def check_param(params):
    keys = ["sensor", "measurement"]
    if not check_keys(params, keys):
        raise ValueError("Param file is bad!")

    keys = ["flag", "sensor_types"]
    if not check_keys(params["sensor"], keys):
        raise ValueError("Param['sensor'] is bad!")

    keys = ["flag", "component"]
    if not check_keys(params["measurement"], keys):
        raise ValueError("Param['measurement'] is bad!")

    print("=" * 10 + " Path info " + "=" * 10)
    pprint(params)


def assert_windows_and_measurements_same_length(windows, measures):
    if len(windows) != len(measures):
        raise ValueError("Length of windows and measures are not the same!")
    for sta in windows:
        if len(windows[sta]) != len(measures[sta]):
            raise ValueError("Length of windows and measures fails at "
                             "station level: %s" % sta)
        for chan in windows[sta]:
            if len(windows[sta][chan]) != len(measures[sta][chan]):
                raise ValueError("Length of windows and measures fails at: "
                                 "channel level: %s" % chan)


def run_window_filter(paths, params, verbose=False):
    check_path(paths)
    check_param(params)

    window_file = paths["window_file"]
    station_file = paths["station_file"]
    output_file = paths["output_file"]
    measurement_file = paths["measurement_file"]

    windows = load_json(window_file)
    # count the number of windows in the original window file
    nchans_old, nwins_old, nwins_comp_old = count_windows(windows)
    stations = load_json(station_file)
    measurements = load_json(measurement_file)

    # filter the window based on given sensor types
    windows_new, measures_new, log = filter_windows(
        windows, stations, measurements, params, verbose=verbose)

    nchans_new, nwins_new, nwins_comp_new = count_windows(windows_new)

    assert_windows_and_measurements_same_length(windows_new, measures_new)
    print("=" * 10 + " Summary " + "=" * 10)
    print("channels: %d --> %d" % (nchans_old, nchans_new))
    print("windows: %d -- > %d" % (nwins_old, nwins_new))
    print("Old component windows: %s" % (nwins_comp_old))
    print("New component windows: %s" % (nwins_comp_new))

    # dump the new windows file to replace the original one
    print("Filtered window files: %s" % output_file)
    dump_json(windows_new, output_file)

    new_measure_file = measurement_file + ".filter"
    print("Filtered measurement file: %s" % new_measure_file)
    dump_json(measures_new, new_measure_file)

    # dump the log file
    logfile = os.path.join(os.path.dirname(output_file), "filter.log")
    print("Log file located at: %s" % logfile)
    dump_json(log, logfile)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', action='store', dest='path_file', required=True,
                        help="path file")
    parser.add_argument('-p', action='store', dest='param_file', required=True,
                        help="param file")
    parser.add_argument('-v', action='store_true', dest='verbose',
                        help="verbose flag")
    args = parser.parse_args()

    paths = load_json(args.path_file)
    params = load_yaml(args.param_file)

    run_window_filter(paths, params, verbose=args.verbose)


if __name__ == "__main__":
    main()



"""

Some important and simple IO help functions.

:copyright:
    Wenjie Lei (lei@princeton.edu) Year? pyasdf
    Lucas Sawade (lsawade@princeton.edu) 2019

:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

"""

from __future__ import (absolute_import, division, print_function)
import json
import yaml


def load_json(filename):
    with open(filename) as fh:
        return json.load(fh)


def dump_json(content, filename):
    with open(filename, 'w') as fh:
        json.dump(content, fh, indent=2, sort_keys=True)


def check_dict_keys(dict_to_check, keys):
    if not isinstance(dict_to_check, dict):
        raise TypeError("input dict_to_check should be type of dict: %s"
                        % (type(dict_to_check)))

    set_input = set(dict_to_check.keys())
    set_stand = set(keys)

    if set_input != set_stand:
        print("More: %s" % (set_input - set_stand))
        print("Missing: %s" % (set_stand - set_input))
        raise ValueError("Keys is not consistent: %s --- %s"
                         % (set_input, set_stand))


def write_yaml_file(d, filename, **kwargs):
    """Writes dictionary to given yaml file.

    Args:
          d: Dictionary to be written into the yaml file
          filename: string with filename of the file to be written.

    """
    with open(filename, 'w') as yaml_file:
        yaml.dump(d, yaml_file, default_flow_style=False, **kwargs)


def read_yaml_file(filename):
    with open(filename, "rb") as fh:
        return yaml.load(fh, Loader=yaml.FullLoader)


def smart_read_yaml(yaml_file, mpi_mode=True, comm=None):
    """
    Read yaml file into python dict, in mpi_mode or not
    """
    if not mpi_mode:
        yaml_dict = read_yaml_file(yaml_file)
    else:
        if comm is None:
            comm = _get_mpi_comm()
        rank = comm.rank
        if rank == 0:
            try:
                yaml_dict = read_yaml_file(yaml_file)
            except Exception as err:
                print("Error in read %s as yaml file: %s" % (yaml_file, err))
                comm.Abort()
        else:
            yaml_dict = None
        yaml_dict = comm.bcast(yaml_dict, root=0)
    return yaml_dict


def is_mpi_env():
    """
    Test if current environment is MPI or not
    """
    try:
        import mpi4py
    except ImportError:
        return False

    try:
        import mpi4py.MPI
    except ImportError:
        return False

    if mpi4py.MPI.COMM_WORLD.size == 1 and mpi4py.MPI.COMM_WORLD.rank == 0:
        return False

    return True


def _get_mpi_comm():
    from mpi4py import MPI
    return MPI.COMM_WORLD



"""

This script contains functions to invert the produced data

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

Last Update: June 2019

"""

import yaml

# CMT3D
from pycmt3d.source import CMTSource
from pycmt3d import Cmt3D
from pycmt3d import DataContainer
from pycmt3d import WeightConfig, Config
from pycmt3d.constant import PARLIST

# Gradient3D

import os
import glob

# Get logger to log progress
from gcmt3d import logger


def read_yaml_file(filename):
    """read yaml file"""
    with open(filename) as fh:
        return yaml.load(fh, Loader=yaml.FullLoader)


def invert(cmt_file_db, param_path):
    """Runs the actual inversion.

    :param cmt_file_db:
    :param param_path:
    :return: Nothing, inversion results are written to file.
    """

    # Load Database Parameters
    databaseparam_path = os.path.join(param_path,
                                      "Database/DatabaseParameters.yml")
    DB_params = read_yaml_file(databaseparam_path)

    # Inversion Params
    inversionparam_path = os.path.join(param_path,
                                       "CMTInversion/InversionParams.yml")
    INV_params = read_yaml_file(inversionparam_path)

    # Get processing path from cmt_filename in database
    cmt_dir = os.path.dirname(os.path.abspath(cmt_file_db))

    # Create cmt source:
    cmtsource = CMTSource.from_CMTSOLUTION_file(cmt_file_db)

    # Inversion dictionary directory
    inv_dict_dir = os.path.join(cmt_dir, "inversion", "inversion_dicts")

    # Inversion dictionaries
    inv_dict_files = glob.glob(os.path.join(inv_dict_dir, "inversion*"))

    # Inversion output directory
    inv_out_dir = os.path.join(cmt_dir,
                               "inversion", "inversion_output", "cmt3d")

    # WRite start of inversion process
    logger.info(" ")
    logger.info("#######################################################")
    logger.info("#                                                     #")
    logger.info("#      Starting CMT3D Inversion ...                   #")
    logger.info("#                                                     #")
    logger.info("#######################################################")
    logger.info(" ")

    # Creating Data container
    dcon = DataContainer(parlist=PARLIST[:DB_params["npar"]])

    for _i, inv_dict_file in enumerate(inv_dict_files):

        # Get processing band
        bandstring1 = str(os.path.basename(inv_dict_file).split(".")[1])
        if "surface" in bandstring1 or "body" in bandstring1:
            bandstring = bandstring1.split("#")[0]
        else:
            bandstring = bandstring1

        band = [float(x) for x in bandstring.split("_")]

        logger.info(" ")
        logger.info("  " + 54 * "*")
        logger.info("  Getting data for inversion from period band:")
        logger.info("  Low: %d s || High: %d s" % tuple(band))
        logger.info("  " + 54 * "*")
        logger.info(" ")

        # Load inversion file dictionary
        inv_dict = read_yaml_file(inv_dict_file)
        asdf_dict = inv_dict["asdf_dict"]
        window_file = inv_dict["window_file"]

        # Adding measurements
        # Print Inversion parameters:
        logger.info("  Adding measurements to data container:")
        logger.info("  _____________________________________________________")
        logger.info(" ")

        # Add measurements from ASDF file and windowfile
        logger.info("  Window file:")
        logger.info("   " + window_file)
        logger.info(" ")
        logger.info("  ASDF files:")
        for key, value in asdf_dict.items():
            logger.info("     " + key + ": " + value)

        dcon.add_measurements_from_asdf(window_file, asdf_dict)

        logger.info("  _____________________________________________________")
        logger.info("  ... ")
        logger.info("  ")
        logger.info("  ... ")

    logger.info("  Setting up inversion classes .... ")
    logger.info("  " + 54 * "*")
    logger.info("  ... ")

    # Setting up weight config
    inv_weight_config = INV_params["weight_config"]

    weight_config = WeightConfig(
        normalize_by_energy=inv_weight_config["normalize_by_energy"],
        normalize_by_category=inv_weight_config["normalize_by_category"],
        azi_bins=inv_weight_config["azi_bins"],
        azi_exp_idx=inv_weight_config["azi_exp_idx"])

    # Setting up general inversion config
    inv_params = INV_params["config"]

    cmt3d_config = Config(
        DB_params["npar"],
        dlocation=float(inv_params["dlocation"]),
        ddepth=float(inv_params["ddepth"]),
        dmoment=float(inv_params["dmoment"]),
        weight_data=bool(inv_params["weight_data"]),
        station_correction=bool(inv_params["station_correction"]),
        zero_trace=bool(inv_params["zero_trace"]),
        double_couple=bool(inv_params["double_couple"]),
        bootstrap=bool(inv_params["bootstrap"]),
        bootstrap_repeat=int(inv_params["bootstrap_repeat"]),
        weight_config=weight_config,
        taper_type=inv_params["taper_type"],
        damping=float(inv_params["damping"]))

    logger.info("  PyCMT3D is finding an improved CMTSOLUTION .... ")
    logger.info("  " + 54 * "*")
    logger.info(" ")
    logger.info(" ")

    # Invert for parameters
    cmt3d = Cmt3D(cmtsource, dcon, cmt3d_config)
    cmt3d.source_inversion()
    cmt3d.compute_new_syn()
    # Create inversion class
    # if bool(INV_params["gridsearch"]):
    #     inv = Inversion(cmtsource, dcon, cmt3d_config, grad3d_config)
    # else:
    #     inv = Inversion(cmtsource, dcon, cmt3d_config, mt_config=None)

    # Run inversion
    if bool(INV_params["statistics_plot"]):
        cmt3d.plot_stats_histogram(outputdir=inv_out_dir)

    # Plot results
    if bool(INV_params["summary_plot"]):
        cmt3d.plot_summary(inv_out_dir, figure_format="pdf")
    #
    # if bool(INV_params["statistics_plot"]):
    #     # Plot Statistics for inversion
    #     inv.G.plot_stats_histogram(outputdir=inv_out_dir)
    #
    if bool(INV_params["summary_json"]):
        cmt3d.write_summary_json(outputdir=inv_out_dir, mode="global")

    if bool(INV_params["write_new_cmt"]):
        cmt3d.write_new_cmtfile(outputdir=inv_out_dir)

    if bool(INV_params["write_new_synt"]):
        cmt3d.write_new_syn(outputdir=os.path.join(inv_out_dir, "new_synt"),
                            file_format="asdf")

    if bool(INV_params["plot_new_synthetics"]):
        cmt3d.plot_new_synt_seismograms(
            outputdir=os.path.join(inv_out_dir, "waveform_plots"),
            figure_format="pdf")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', action='store', dest='cmt_file',
                        required=True, help="Path to CMT file in database")
    parser.add_argument('-p', action='store', dest='param_path', required=True,
                        help="Path to Parameter Directory")
    args = parser.parse_args()

    invert(args.cmt_file, args.param_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
This script extract channel instrument information for one asdf file
and store in json file. This will save up time for re-parsing the
stationxml file.
'''

from __future__ import print_function, division, absolute_import
import argparse
from gcmt3d.asdf.utils import load_json, dump_json
from gcmt3d.asdf.stations import extract_station_info_from_asdf


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', action='store', dest='path_file', required=True,
                        help="path file")
    parser.add_argument('-v', action='store_true', dest='verbose',
                        help="verbose flag")
    args = parser.parse_args()

    paths = load_json(args.path_file)
    input_asdf = paths["input_asdf"]
    outputfn = paths["outputfile"]

    print("input asdf: %s" % input_asdf)
    print("output sensors json: %s" % outputfn)

    asdf_sensors = extract_station_info_from_asdf(
        input_asdf, verbose=args.verbose)

    dump_json(asdf_sensors, outputfn)


if __name__ == "__main__":

    main()



'''

This script/set of functions is used to create a set of sources for parallel
specfem simulations

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

Last Update: January 2020

'''
import pycmt3d


class GCMT3D(object):

    def __init__(self, dcon: pycmt3d.data_container.DataContainer,
                 cmt3d_config: pycmt3d.Config,
                 grid3d_config: pycmt3d.grid3d.Grid3dConfig,
                 weight_config: pycmt3d.config.WeightConfig,
                 npar=9):

        pass



"""

This file contains functions and classes to perform forward and adjoint
modelling of waveforms

:copyright:
    Lucas Sawade (lsawade@princeton.edu, 2019)

:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)

"""

import numpy as np
from numpy.fft import fft, ifft, fftfreq, fftshift

import matplotlib.pyplot as plt


def gaussian(t, to, sig, amp):
    """Create Gaussian pulse.

    :param t: time vector
    :param to: time in the center
    :param sig: standard deviation
    :param amp: amplitude
    :return: vector with same size of t containing the corresponding pulse

    """
    return amp * np.exp(-0.5 * (t-to)**2 / (sig * sig))


def bimodal_gaussian(t, amp1=2.5, to1=20, sig1=5, amp2=1.75, to2=32, sig2=3):
    """ Creates bimodal gaussian from gaussian function.

    :param t: time vector
    :param amp1: amplitude of first gaussian
    :param to1: center time of first gaussian
    :param sig1: standard deviation of first gaussian
    :param amp2: amplitude of second gaussian
    :param to2: center time of second gaussian
    :param sig2: standard deviation of second gaussian
    :return:
    """

    g1 = gaussian(t, to1, sig1, amp1)
    g2 = gaussian(t, to2, sig2, amp2)

    return g1 + g2


def single_greens_function(t, dt, delta_loc_t, delta_amp):
    """Create a single, simple Green's function from time locations and
    amplitudes of delta functions.

    :param t: time vector
    :param dt: sampling time
    :param delta_loc_t: time locations of Green's function peaks
    :param delta_amp: corresponding amplitudes
    :return: Green's function corresponding to time vector t

    """

    # Create empty Green's function
    green = np.zeros(len(t))

    # Populate it using the list
    for _i, (delta_t, delta_a) in enumerate(zip(delta_loc_t, delta_amp)):
        it = np.argmin(np.abs(delta_t-t))
        green[it] = delta_a

    return green


def synthetic_seismogram(green, wavelet):
    """Takes in a Gaussian and a Green's function and computes the
    seismogram.

    :param green: Green's function
    :param wavelet: source wavelet
    :return: seismogram
    """
    return np.real(ifft(fft(wavelet) * fft(green)))


def noisy_seismogram(t, seismogram, noise_amp=5):
    """ Takes in t and a seismogram to compute the

    :param t:
    :param seismogram:
    :return:
    """

    # Noise
    signoise = 2 * np.sqrt(3)

    # Create filter to take out high frequency noise
    filtgauss = gaussian(t, 75, signoise, 1.)
    filtgauss = filtgauss / sum(filtgauss)

    # Amplitude for the noise
    amp = noise_amp
    noise = 2 * (np.random.uniform(size=t.shape) - 0.5) * amp

    # Compute filtered noise
    filtnoise = np.real(ifft(fft(noise) * fft(filtgauss)))

    # Add noise to original seismogram
    noisemogram = seismogram + filtnoise

    return noisemogram


def synthetic_traces(nr=20, dx=10, nt=3001, dt=0.01, green_vel=None,
                     delta_loc_t=None, delta_amp=None,
                     amp1=2.5, to1=20, sig1=5, amp2=1.75, to2=32, sig2=3,
                     noise_amp=5):
    """
    :param nr: Number of receivers
    :param dx: receiver spacing
    :param nt: number of samples
    :param dt: sampling interval
    :param green_vel: green's function velocities between the peaks
                      Note that green_vel, delta_loc_t, delta_amp have to
                      have the same number of elements
    :param delta_loc_t: initial locations of the green's function things
    :param delta_amp: corresponding amplitudes
    :param amp1: amplitude of first gaussian
    :param to1: center time of first gaussian
    :param sig1: standard deviation of first gaussian
    :param amp2: amplitude of second gaussian
    :param to2: center time of second gaussian
    :param sig2: standard deviation of second gaussian
    :param noise_amp: amplitude of the noise, the noise is filtered with a
                      gaussian to avoid high frequency noise.
    :return: time vector, space vector and set of green's functions, synthetic
    traces and their
    noisified
             corresponding observed data --> (t, xr, Garray, Sarray, Oarray)
    """

    # Time parameters
    t = np.arange(nt)*dt

    # Multiple seismograms
    obs = np.zeros((nr, nt))
    syn = np.zeros((nr, nt))
    G = np.zeros((nr, nt))
    xr = []

    # Source-time function
    wavelet = bimodal_gaussian(t, amp1, to1, sig1, amp2, to2, sig2)

    # iterate over receivers
    for ir in range(nr):

        # Receiver distance from 0
        xr.append(ir * dx)

        # Compute velocity dependent time shift for Green's functions
        tshift = xr[ir] / green_vel

        # Change locations for initial location of the delta functions in trace
        delta_loc_t_tmp = delta_loc_t + tshift

        # Compute green's function using artificial damping
        idecay = np.arange(delta_amp.shape[0])
        delta_amp_tmp = delta_amp * (1-0.99*idecay/nr)

        # Compute green's functions
        G[ir, :] = single_greens_function(t, dt,
                                          delta_loc_t_tmp,
                                          delta_amp_tmp)

        # Compute seismograms
        syn[ir, :] = np.real(ifft(fft(wavelet)*fft(G[ir, :])))

        # Compute noisy, 'observed' seismograms
        obs[ir, :] = noisy_seismogram(t, syn[ir, :], noise_amp=noise_amp)

    return t, np.array(xr), G, syn, obs, wavelet


def plot_one_trace_set(t, wavelet, G, syn, obs):
    """Plots synthetic data.
    :param t: time vector
    :param wavelet: stf
    :param G: green's function
    :param syn: synthetic
    :param obs: observed
    :return:
    """

    fig, ax = plt.subplots(3, 1, sharex=True, sharey=True, figsize=(10, 6))

    # True STF
    ax[0].plot(t, wavelet)
    ax[0].set_xlabel('Time in s')
    ax[0].set_ylabel('Amplitude')
    ax[0].set_title('True source wavelet')

    # Green function
    ax[1].plot(t, G)
    ax[1].set_xlabel('Time in s')
    ax[1].set_ylabel('Amplitude')
    ax[1].set_title('Green function')

    # Seismograms
    ax[2].plot(t, syn, label='Synthetic')
    ax[2].plot(t, obs, label='Observed')
    ax[2].set_xlabel('Time in s')
    ax[2].set_ylabel('Amplitude')
    ax[2].set_title('Seismograms')
    plt.legend(loc='best')
    ax[2].set_xlim([np.min(t), np.max(t)])

    plt.tight_layout()

    plt.show(block=False)


def plot_section(t, syn, obs):
    """Plots section of data of the synthetic and observed data"""

    fig, ax = plt.subplots(1, 2, sharex=True, sharey=True, figsize=(8, 8))

    for ir in range(syn.shape[0]):

        # Synthetic
        ax[0].plot(t, ir + syn[ir, :], 'k')
        ax[0].set_xlabel('Time in s')
        ax[0].set_ylabel('Amplitude')
        ax[0].set_title('Synthetic')

        # Noisy observed data
        ax[1].plot(t, ir + obs[ir, :], 'k')
        ax[1].set_xlabel('Time in s')
        ax[1].set_title('Observed')
        ax[1].set_xlim([np.min(t), np.max(t)])
        ax[1].set_ylim([-1, syn.shape[0]+1.5])

    plt.tight_layout()

    plt.show(block=False)


def compute_synth(green, src):
    """ Convolution of set of Green's functions

    :param green: Green's function
    :param src:
    :return:
    """
    # Get frequency spectrum of source time function
    SRC = fft(src)

    # Get frequency spectra of Green's functions
    GRE = fft(green, axis=1)

    # Convolve the two and return matrix containing the synthetic
    syn = np.real(ifft(GRE*SRC, axis=1))

    return syn


def deconvolution(obs, green, lambd):
    """ Takes in the observed data and the green's function to obtain the
    source wavelet estimate.

    :param obs:
    :param green:
    :return:
    """

    nr, nt = obs.shape
    num = np.zeros(nt)
    den = np.zeros(nt)

    for ir in range(len(obs)):

        OBS = fft(obs[ir, :])
        GRE = fft(green[ir, :])

        # Sum all
        num = num + np.conj(GRE) * OBS
        den = den + np.conj(GRE) * GRE

    # Get maximum value of denominator
    maxden = np.max(np.abs(den))

    # Waterlevel
    wl = lambd * maxden

    # Deconvolution using the waterlevel
    src = np.real(ifft(num / (den+wl).T))

    # Compute fit to original data
    res = obs
    chi0 = 0.5 * np.sum(np.sum(res ** 2))

    syn = compute_synth(green, src)
    res = obs - syn
    chi = 0.5 * np.sum(np.sum(res ** 2))

    print(chi/chi0)

    return src, syn


def plot_source_comparison(t, wavelet, inversions, labels):
    """

    :param t: time vector
    :param wavelet: original wavelet
    :param inversions: list of inverted
    :params labels: labels corresponding to the inverted stfs
    :return:
    """

    plt.figure(figsize=(10, 6))
    plt.plot(t, wavelet, label="Original")
    for stf, label in zip(inversions, labels):
        plt.plot(t, stf, label=label)
    plt.legend()
    plt.show(block=False)


def landweber(obs, G, dt, maxT, crit, lamb, type="2"):
    """

    :param obs: observed traces
    :param G: Green's functions
    :param maxT: time after which STF is forced to zero
    :param crit: critical value for stopping the iteration
    :param dt: time sampling
    :param lamb: waterlevel for deconvolution if type 2 is chosen. Unused if
                 type is "1".
    :param type: string defining the type of landweber method. Type 1 is the
                 method using the steepest decent; type 2 is using a Newton
                 step.
    :return:
    """

    if type == "1":
        compute_gradient = compute_gradient_sd
    elif type == "2":
        compute_gradient = compute_gradient_newton
    else:
        raise ValueError("Wrong type!")

    # Get informations about size and initialize src
    nr, nt = obs.shape
    src = np.zeros(nt)

    # Compute objective function and gradient
    syn = compute_synth(G, src)
    res = obs - syn
    chi = 0.5 * np.sum(np.sum(res ** 2))
    chi0 = chi
    grad, alpha = compute_gradient(res, G, lamb)

    # Manage windowing tapered window in the future? e.g., tukey?
    itstop = int(np.floor(maxT / dt) + 1)
    Ptime = np.ones(nt)
    Ptime[itstop + 1: itstop + 31] = np.arange(30, 0, -1) / 30.
    Ptime[itstop + 31:] = 0.

    # Perform iterative deconvolution (inverse problem)
    it = 1
    perc = 0.05
    chip = chi0
    # llb = 0.1

    # Initialize list to save iterations
    src_list = []
    chi_list = []

    while chi > crit * chi0 and it <= nt:

        # Regularized gradient
        gradreg = grad

        if type == "1":
            srctmp = src + gradreg
        else:
            srctmp = src + perc * gradreg

        # Window source time function --> zero after some time T
        srctmp = srctmp * Ptime

        # Enforce positivity
        srctmp[np.where(srctmp < 0)[0]] = 0

        # Compute misfit function and gradient
        syn = compute_synth(G, srctmp)
        res = obs - syn
        chi = 0.5 * np.sum(np.sum(res ** 2))
        grad, _ = compute_gradient(res, G, lamb)
        it = it + 1

        # Check convergence
        if chi > chip:
            print("NOT CONVERGING")
            break

        # Update
        # chi / chi0
        chip = chi
        src = srctmp

        chi_list.append(chi)
        src_list.append(src)

    # Final misfit function
    print(chi / chi0)
    print(it)

    return src, src_list, chi_list


def compute_gradient_newton(resid, green, lamb):
    """ Compute Gradient using the waterlevel deconvolution which computes
    the Newton Step.

    :param resid: residual
    :param green: green's function
    :param lamb: waterlevel scaling
    :return:
    """
    # Get infos
    nr, nt = green.shape

    # FFT of residuals and green functions
    RES = fft(resid, axis=1)
    GRE = fft(green, axis=1)

    # Compute gradient (full wavelet estimation)
    num = np.sum(RES * np.conj(GRE), axis=0)
    den = np.sum(GRE * np.conj(GRE), axis=0)

    # Waterlevel?
    wl = lamb * np.max(np.abs(den))
    grad = np.real(ifft(num / (den + wl)))

    # Step value
    hmax = 1

    return grad, hmax


def compute_gradient_sd(resid, green, lamb):
    """ Compute the Gradient using the steepest decent method
    :param resid:
    :param green:
    :return:
    """

    # FFT of residuals and green functions
    RES = fft(resid, axis=1)
    GRE = fft(green, axis=1)

    # Compute gradient (full wavelet estimation)
    num = np.sum(RES * np.conj(GRE), axis=0)
    den = np.sum(GRE * np.conj(GRE), axis=0)

    # Waterlevel?
    tau = 1/np.max(np.abs(den))
    grad = np.real(ifft(num * tau))

    # Step value
    hmax = 1

    return grad, hmax


def plot_comparison_section(t, obs, syn, syn_decon, syn_lw):
    """Plot section to compare convolution results

    :param t: time vector
    :param obs: observed data
    :param syn_decon: synthetics using waterlevel decon
    :param syn_lw: synthetics using landweber iterative.
    :return: plot
    """

    fig, ax = plt.subplots(1, 2, sharex=True, sharey=True, figsize=(8, 8))

    for ir in range(obs.shape[0]):
        # Observed + Standard deconvolution
        if ir == 0:
            label_obs = 'Obs'
            label_syn = 'Syn'
            label_syn_decon = 'Inv'
        else:
            label_obs = None
            label_syn = None
            label_syn_decon = None

        ax[0].plot(t, ir + obs[ir, :], 'k', label=label_obs)
        ax[0].plot(t, ir + syn[ir, :], c=(0.85, 0.2, 0.2), label=label_syn)
        ax[0].plot(t, ir + syn_decon[ir, :], c=(0.2, 0.2, 0.85),
                   label=label_syn_decon)
        ax[0].set_xlabel('Time in [s]')
        ax[0].set_ylabel('Amplitude')
        ax[0].set_title('Waterlevel')

        # Observed + Landweber iterative deconvolution
        ax[1].plot(t, ir + obs[ir, :], 'k')
        ax[1].plot(t, ir + syn[ir, :], c=(0.85, 0.2, 0.2))
        ax[1].plot(t, ir + syn_lw[ir, :], c=(0.2, 0.2, 0.85))
        ax[1].set_xlabel('Time in [s]')
        ax[1].set_title('Landweber')
        ax[1].set_xlim([np.min(t), np.max(t)])
        ax[1].set_ylim([-1, syn_lw.shape[0] + 1.5])

    plt.figlegend(frameon=True, fancybox=False, loc='lower center', ncol=3,
                  framealpha=1, edgecolor='k', facecolor='w',
                  bbox_to_anchor=(0.52, 0.925))
    plt.tight_layout()

    plt.show(block=False)


def plot_stf_evolution(t, stf, stf_list, skip=5):

    fig, ax = plt.subplots(1, 2, figsize=(10, 5))

    # STF plotting
    for _i, f in enumerate(stf_list[::skip]):
        ax[0].plot(t, f,
                   c=(0.25, 0.25, 0.25, 1),
                   alpha=((len(stf_list[::skip]) * 0.05 + _i)
                          / (len(stf_list[::skip]) * 2)),
                   label=None)
    ax[0].plot(t, stf_list[-1], c=(0.8, 0.2, 0.2, 1), label="Inversion")
    ax[0].plot(t, stf, c=(0.2, 0.2, 0.8, 1), label="Original")
    # ax[0].set_xlim([iT, fT])
    ax[0].set_xlabel('Time in [s]')
    plt.legend()

    # Frequency spectrum
    freq = fftshift(fftfreq(len(stf_list[0]), (t[1]-t[0])))
    for _i, f in enumerate(stf_list[::skip]):
        ax[1].plot(freq, np.real(np.abs(fftshift(fft(f)))),
                   c=(0.25, 0.25, 0.25, 1),
                   alpha=((len(stf_list[::skip])*0.05+_i)
                          / (len(stf_list[::skip]) * 2)))

    ax[1].plot(freq, np.real(np.abs(fftshift(fft(stf_list[-1])))),
               c=(0.8, 0.2, 0.2, 1), label="Inversion")
    ax[1].plot(freq, np.real(np.abs(fftshift(fft(stf)))),
               c=(0.2, 0.2, 0.8, 1),
               label="Original")
    plt.legend()
    ax[1].set_xlim([0, .5*np.max(1/(t[1]-t[0]))])
    ax[1].set_ylim([0, np.max(stf_list)])
    ax[1].set_xlabel('Frequency in [Hz]')
    plt.tight_layout()

    plt.show(block=False)

    return ax
